/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Diorite{
	
	/**
	 * File storage abstraction
	 */
	public class Storage : Object{
		protected string path_name;
		
		[Description(nick = "User data dir", blurb = "Directory used to store user-specific data. Read-write access may be available.")]
		public GLib.File? user_data_dir{
			owned get{
				var dir = File.new_for_path(
					Environment.get_user_data_dir()).get_child(path_name);
				if(dir.query_file_type(0) == FileType.DIRECTORY){
					return dir;
				}
				return null;
			}
		}
		
		[Description(nick = "System data dirs", blurb = "Directories with system-wide data. Read-only access.")]
		public GLib.File[] data_dirs{
			owned get{
				File[] dirs = {};
				foreach(string path in Environment.get_system_data_dirs()){
					var dir = File.new_for_path(path).get_child(path_name);
					if(dir.query_file_type(0) == FileType.DIRECTORY){
						dirs += dir;
					}
				}
				return dirs;
			}
		}
		
		[Description(nick = "User cache dir", blurb = "Directory used to store user-specific cached data. Read-write access may be available.")]
		public GLib.File? user_cache_dir{
			owned get{
				return File.new_for_path(
					Environment.get_user_cache_dir()).get_child(path_name);
			}
		}
		
		[Description(nick = "User configuration dir", blurb = "Directory used to store user-specific configuration data. Read-write access may be available.")]
		public GLib.File? user_config_dir{
			owned get{
				return File.new_for_path(
					Environment.get_user_config_dir()).get_child(path_name);
			}
		}
		
		/**
		 * Creates new storage identified by the id
		 * 
		 * @param id identifier of the storage
		 */
		public Storage(string id){
			this.path_name = id;
		}
		
		/**
		 * Returns new child storage.
		 * 
		 * @param id	child storage id
		 * @return 		child storage 
		 */
		public Storage get_child(string id){
			return new Storage(path_name + "/" + id);
		}
		
		/**
		 * Returns the default path of configuration file with given name
		 * 
		 * @param file Name of configuration file
		 * @return default configuration file
		 */
		public File? get_default_config_file(string file){
			return File.new_for_path(Environment.get_user_config_dir())
			.get_child(path_name).get_child(file);
		}
		
		/**
		 * Returns the default path of cache file with given name
		 * 
		 * @param file Name of cache file
		 * @return default cache file
		 */
		public File? get_default_cache_file(string file){
			return File.new_for_path(Environment.get_user_cache_dir())
			.get_child(path_name).get_child(file);
		}
		
		/**
		 * Returns the default path of cache directory or a subdirectory with given name
		 * 
		 * @param subdir	name of a cache subdirectory or null for cache root directory
		 * @return default cache directory
		 */
		public File get_default_cache_dir(string? subdir=null){
			var dir = File.new_for_path(Environment.get_user_cache_dir()).get_child(this.path_name);
			if(subdir != null){
				return dir.get_child(subdir);
			}
			return dir;
		}
		
		/**
		 * Returns the default path of cache dir with given name
		 * 
		 * @param dir Name of data directory
		 * @return default cache file
		 */
		public File get_default_data_dir(string dir){
			return File.new_for_path(Environment.get_user_data_dir())
			.get_child(this.path_name).get_child(dir);
		}
		
		public File? get_data_file(string name){
			File f = File.new_for_path(Environment.get_user_data_dir()).get_child(path_name + "/" + name);
			if(f.query_file_type(0) == FileType.REGULAR) return f;
			foreach(string path in Environment.get_system_data_dirs()){
				f = File.new_for_path(path).get_child(path_name + "/" + name);
				if(f.query_file_type(0) == FileType.REGULAR) return f;
			}
			return null;
		}
	}
} // namespace Diorite
