/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite;
using Diorite.Logger;
using Diorite.System;
using Nuvola.Settings;
using Nuvola.Actions;

namespace Nuvola{

private const string APP_CRASHED = "crashed";
private const string APP_CHECK_CRASHED = "check_crashed";
private const string LAST_SERVICE_KEY = "last_service";

namespace Actions{
	public const string PURGE_COOKIES = "purge-cookies";
	public const string PURGE_CACHE = "purge-cache";
	public const string PURGE_SETTINGS = "purge-settings";
	public const string PURGE_UI_STATE = "purge-ui-state";
	public const string DISABLE_PROXY = "disable-proxy";
}

/**
 * Struct containing command line arguments. Check Args.options for their meaning.
 */
struct Args {
	static bool debug;
	static bool verbose;
	static bool version;
	static bool time;
	static bool autoplay;
	static bool noautoplay;
	static bool debug_javascript;
	static bool debug_navigation;
	static bool debug_http;
	static bool no_web_plugins;
	
	public static const OptionEntry[] options = {
		{ "version", 'V', 0, GLib.OptionArg.NONE, ref Args.version, "Show version of program", null },
		{ "verbose", 'v', 0, GLib.OptionArg.NONE, ref Args.verbose, "Print information to terminal", null },
		{ "debug", 'D', 0, GLib.OptionArg.NONE, ref Args.debug, "Print debug messages", null },
		{ "debug-javascript", 0, 0, GLib.OptionArg.NONE, ref Args.debug_javascript, "Enable JavaScript debugging.", null },
		{ "debug-navigation", 0, 0, GLib.OptionArg.NONE, ref Args.debug_navigation, "Enable navigation debugging.", null },
		{ "debug-http", 0, 0, GLib.OptionArg.NONE, ref Args.debug_http, "Enable debugging of HTTP communication.", null },
		{ "time", 'T', 0, GLib.OptionArg.NONE, ref Args.time, "Show time in debug messages", null },
		{ "auto-play", 0, 0, GLib.OptionArg.NONE, ref Args.autoplay, "Automatically start playback (if possible). Overrides configuration option.", null },
		{ "no-auto-play", 0, 0, GLib.OptionArg.NONE, ref Args.noautoplay, "Do not automatically start playback. Overrides configuration option.", null },
		{ "no-web-plugins", 0, 0, GLib.OptionArg.NONE, ref Args.no_web_plugins, "Disables web plugins.", null },
		{ null }
	};
}

/**
 * Main entry point
 */
int main(string[] args){
	var start_time = TimeVal();
	start_time.get_current_time();
	Gtk.init(ref args);
	// Save command and arguments
	var command = string.joinv(" ", args);
	
	/* Parse command line arguments */
	var opt_context = new OptionContext("");
	try{
		opt_context.set_help_enabled(true);
		opt_context.add_main_entries(Args.options, null);
		opt_context.parse(ref args);
	}
	catch(OptionError e){
		/// Shown if unsupported command line argument is used
		stderr.printf(_("Error: Option parsing failed:") + " %s\n\n%s\n",
		e.message, opt_context.get_help(true, null));
	return 1;
	}
	
	// Show version and exit
	if(Args.version){
		stdout.printf("%s %s\n", DISPLAY_NAME, VERSION);
		stdout.printf("Revision: %s, %s\n", REVISION, REVISION_ID);
		return 0;
	}
	
	/* Logger initialization.
	 * Use Diorite.debug() instead of GLib.debug() etc. to use "Nuvola" domain.
	 */
	Diorite.Logger.init("Nuvola", Args.debug ? LogLevelFlags.LEVEL_DEBUG
		: (Args.verbose ? LogLevelFlags.LEVEL_INFO : LogLevelFlags.LEVEL_WARNING));
	if(Args.time){
		Diorite.Logger.time = start_time.tv_sec;
	}
	
	Environment.set_variable("XLIB_SKIP_ARGB_VISUALS", "1", false);
	
	#if STABLE 
		message("Welcome to %s, version: %s", DISPLAY_NAME, VERSION);
	#else
		message("Welcome to %s, development version: %s", DISPLAY_NAME, VERSION);
	#endif
	message("Revision: %s, %s", REVISION, REVISION_ID);
	message("Report any issues/bugs you might find to %s", BUG_TRACKER);
	debug("command: %s", command);
	
	debug("Enabled features: %s", ENABLED_FEATURES);
	debug("Disabled features: %s", DISABLED_FEATURES);
	
	/* Init translations */
	Intl.textdomain(APPNAME);
	Intl.bindtextdomain(APPNAME, LOCALES_DIR);
	
	var app = new Main();
	return app.run();
}

/**
 * The main application class manages operations related to application life cycle
 * (start → load configuration and ui state → initialize objects → save ui state
 * → quit) and loads service object that will be controlled by Controller.
 */
private class Main: Diorite.Application {
	private MainWindow main_window = null;
	private MainWebView main_view;
	private Player player;
	private Service? service = null;
	private Connection connection;
	private ServicesManager services_manager;
	private ObjectContainer objects;
	private Configuration config;
	private MapFile ui_state = null;
	private ExtensionsManager extensions;
	private Storage storage;
	private ServicesManagerView? selector = null;
	private Diorite.Actions actions;
	private Gtk.UIManager ui_manager;
	private PreferencesWindow? preferences_window = null;
	private bool loaded = false;

	/**
	 * Constructor for the main application class
	 * 
	 * @param config Configuration
	 * @param ui_state User interface state
	 * @param storage File storage
	 */
	public Main(){
		base(UNIQUE_NAME, DESKTOP_ENTRY, APPNAME, null);
		icon = APP_ICON;
		version = VERSION;
		display_name = DISPLAY_NAME;
		app_name = APPNAME;
		
		activate.connect(activated);
	}
	
	#if DEBUG_MEMORY
	~Main(){
		debug("~Main");
	}
	#endif
	
	/**
	 * Called when application is about to be activated.
	 * Initializes new instance or activated already
	 * running instance.
	 */
	private void activated(){
		debug("Application activated");
		if(main_window == null){
			shutdown.connect(on_shutdown);
			shutdown.connect_after(on_late_shutdown);
			start();
		}
		else{
			main_window.present();
		}
	}
	
	/**
	 * Start application
	 */
	private void start(){
		ui_manager = new Gtk.UIManager();
		actions = new Diorite.Actions(ui_manager, app_name);
		storage = new Storage(APPNAME);
		config = new ConfigurationFile(storage.get_default_config_file(CONFIG_FILE));
		ui_state = new MapFile(storage.get_default_config_file(UI_STATE_FILE));
		services_manager = new ServicesManager(storage.get_child(SERVICES_DIRNAME));
		var logging = Args.debug_http ? Soup.LoggerLogLevel.HEADERS : Soup.LoggerLogLevel.NONE;
		connection = new Connection.with_session(WebKit.get_default_session(), storage, logging);
		player = new Player(this, config, storage, connection, actions);
		config.value_changed.connect(on_settings_changed);
		player.song_changed.connect(on_song_changed);
		
		investigate_plugins();
		append_ui_actions();
		check_crashed();
		setup_connection();
		prepare_gui();
		load_extensions();
		loaded = true;
		welcome();
	}
	
	/**
	 * Shows crash dialog if a previous instance was not terminated properly.
	 */
	private void check_crashed(){
		if(!this.ui_state.get_bool(APP_CHECK_CRASHED, true)){
			debug("Crash dialog disabled.");
		}
		else if(ui_state.get_bool(APP_CRASHED, false)){
			debug("Crash dialog");
			var crash = new CrashDialog(actions, this.connection.is_soup_buggy());
			var response = crash.run();
			if(response == Gtk.ResponseType.OK){
				crash.run_actions();
				if(crash.disable_check.active){
					this.ui_state.set_bool(APP_CHECK_CRASHED, false);
				}
			}
			crash.destroy();
		}
		
		/* Set crash flag to true. Will be set to false on a successful shutdown */
		this.ui_state.set_bool(APP_CRASHED, true);
		try{
			this.ui_state.save();
		}
		catch(Error e){
			critical("Unable to save UI state: %s", e.message);
		}
	}
	
	private void investigate_plugins(){
		const string PLUGIN_DIR_KEY = "MOZ_PLUGIN_PATH";
		const string NUVOLA_FLASH_DIR = "/opt/nuvolaplayer/flash/wrapped";
		const string NUVOLA_FLASH_LIBNAME = "/npwrapper.libflashplayer.so";
		
		var plugin_dirs = Environment.get_variable(PLUGIN_DIR_KEY);
		if (plugin_dirs == null || plugin_dirs == ""){
			plugin_dirs = NUVOLA_FLASH_DIR;
		}
		else{
			plugin_dirs = "%s:%s".printf(NUVOLA_FLASH_DIR, plugin_dirs);
		}
		Environment.set_variable(PLUGIN_DIR_KEY, plugin_dirs, true);
		
		var db = WebKit.get_web_plugin_database();
		var plugins = db.get_plugins();
		var compatible_flash_available = false;
		
		// Try to find compatible Flash
		foreach(var plugin in plugins){
			var path = plugin.get_path();
			if(path.has_suffix(NUVOLA_FLASH_LIBNAME)){
				compatible_flash_available = true;
				debug("Compatible Flash: %s, %s", plugin.get_name(), path);
				break;
			}
			
			debug("WebPlugin: %s, %s", plugin.get_name(), path);
		}
		
		// Compatible Flash found, disable other plugins
		if(compatible_flash_available){
			foreach(var plugin in plugins){
				var path = plugin.get_path();
				if(!path.has_suffix(NUVOLA_FLASH_LIBNAME)){
					plugin.enabled = false;
				}
			}
		}
		
		services_manager.compatible_flash_plugin = compatible_flash_available;
	}
	
	/**
	 * Adds user interface actions related to main application object
	 */
	private void append_ui_actions(){
		actions.add_actions({
			/// Menu Service
			new NormalAction("service", null, _("_Service"), null, null),
			new NormalAction("quit", Gtk.Stock.QUIT, null, "<ctrl>Q", null, quit),
			/// Menu item to switch service
			new NormalAction("switch-service", "nuvolaplayer", _("_Switch Service"), "<ctrl>S", _("Switch service"), select_service),
			/// Menu Control
			new NormalAction("control", null, _("_Control"), null, null),
			/// Menu View
			new NormalAction("view",null, _("_View"), null, null),
			/// Action label
			new NormalAction(PURGE_COOKIES, null, _("Remove cookies"), null, null, purge_cookies),
			/// Action label
			new NormalAction(PURGE_CACHE, null, _("Remove files from cache"), null, null, purge_cache),
			/// Action label
			new NormalAction(PURGE_UI_STATE, null, _("Reset state of user interface"), null, null, purge_ui_state),
			/// Action label
			new NormalAction(PURGE_SETTINGS, null, _("Reset settings"), null, null, purge_settings),
			/// Action label
			new NormalAction(DISABLE_PROXY, null, _("Disable HTTP proxy"), null, null, disable_proxy),
			new NormalAction("preferences", Gtk.Stock.PREFERENCES, null, null, null, edit_settings),
			/// Action label
			new NormalAction("service-settings", null, _("Service settings"), null, null, edit_service),
			/// Action label
			new NormalAction("extensions-settings", null, _("Extensions settings"), null, null, edit_extensions),
			/// Menu Help
			new NormalAction("help", null, _("_Help"), "F1", null),
			/// Menu item to open online help
			new NormalAction("questions", Gtk.Stock.HELP, _("Get help online"), null, null, open_support),
			new NormalAction("about", Gtk.Stock.ABOUT, null, null, null, about),
			/// Menu item to open release notes
			new NormalAction("website", Gtk.Stock.INFO, _("Release notes"), null, null, open_release_notes)
		});
	}
	
	/**
	 * Set up network settings (proxy, cache)
	 */
	private void setup_connection(){
		connection.set_up_cookies();
		#if WEB_CACHE
			if(config.data_cache){
				connection.enable_cache(config.data_cache_size);
			}
		#endif
		
		try{
			connection.set_up_proxy(config.proxy_settings,
				config.proxy_type, config.proxy_server,
				config.proxy_port);
		}
		catch(ProxyError e){
			critical("Unable to set up proxy: %s", e.message);
		}
		WebKit.set_cache_model(WebKit.CacheModel.DOCUMENT_VIEWER);
	}
	
	/**
	 * Load main graphical user interface
	 */
	private void prepare_gui(){
		main_window = new MainWindow(this, ui_manager, actions, storage, ui_state);
		main_window.close_action.connect_after(on_close_action);
		
		main_view = new MainWebView(storage.get_child("webkit"), actions);
		main_view.navigation_approved.connect(on_navigation_approved);
		
		main_window.content.add(main_view);
		main_window.load_actions_ui();
		main_view.grab_focus();
		update_window_title();
		main_window.show_all();
	}
	
	/**
	 * Prepares resources for extensions and loads extensions.
	 */
	private void load_extensions(){
		var objects = new HashTable<string, Object>(str_hash, str_equal);
		objects.insert("application", this);
		objects.insert("main_window", main_window);
		objects.insert("main_view", main_view);
		objects.insert("player", player);
		objects.insert("actions", actions);
		objects.insert("storage", storage);
		objects.insert("config", config);
		objects.insert("ui_state", ui_state);
		this.objects = new ObjectContainer(objects);
		
		extensions = new ExtensionsManager(ui_manager, main_window, this.objects);
		foreach(var extension in config.array("extensions")){
			extensions.load(extension);
		}
	}
	
	/**
	 * Shows welcome message when user runs current series for the first time
	 */
	private void welcome(){
		var prev_release = ui_state["release"] ?? "0.1";
		if(prev_release != SERIES){
			ui_state["release"] = SERIES;
			var welcome_box = new Gtk.HBox(false, 5);
			var welcome_label = new Gtk.Label(AboutDialog.get_welcome_message());
			welcome_label.selectable = true;
			welcome_label.use_markup = true;
			welcome_label.can_focus = false;
			welcome_label.xpad = 10;
			welcome_label.ypad = 5;
			welcome_label.wrap_mode = Pango.WrapMode.WORD;
			welcome_label.single_line_mode = false;
			welcome_box.pack_start(welcome_label, true);
			var welcome_close = new Gtk.Button();
			var close_img = new Gtk.Image.from_stock(Gtk.Stock.CLOSE, Gtk.IconSize.BUTTON);
			welcome_close.add(close_img);
			welcome_close.clicked.connect(() => {
				welcome_box.destroy();
			});
			welcome_box.pack_start(welcome_close, false);
			main_window.box.pack_end(welcome_box, false);
			welcome_box.show_all();
		}
		start_or_select_service();
	}
	
	/**
	 * Starts last service or shows service selector
	 */
	private void start_or_select_service(){
		ui_state["release"] = SERIES;
		ui_state["dev_channel"] = DEV_STAGE;
		var service_id = ui_state[LAST_SERVICE_KEY];
		var new_service = services_manager.get_service(service_id);
		if(new_service != null){
			load_service(new_service);
		}
		else{
			select_service();
		}
	}
	
	/**
	 * Loads service
	 */
	private void load_service(Service new_service){
		service = new_service;
		player.service = new_service;
		var js_api = new JSApi(storage, main_view, service, actions);
		player.js_api = js_api;
		main_view.service = new_service;
		main_view.enable_navigation = true;
		var uri = ui_state[Service.CONFIG_PREFIX + "." + new_service.id + ".last_uri"]
		?? new_service.home_page;
		debug("Starting web interface for %s: %s", service.id, uri);
		js_api.load_finished.connect_after(on_service_ready);
		var settings =  main_view.get_settings();
		settings.enable_plugins = !Args.no_web_plugins && service.flash_plugin;
		main_view.set_settings(settings);
		main_view.open(uri);
	}
	
	/**
	 * Show service selector
	 */
	public void select_service(){
		if(selector != null){
			if(service != null){
				destroy_selector();
			}
			return;
		}
		
		if(preferences_window != null){
			// Preferences contains service-specific settings!
			preferences_window.close();
		}
		
		string? prev = service != null ? service.id : null;
		debug("Select service, prev = %s", prev ?? "(none)");
		var template = storage.get_data_file("html/Service.html");
		selector = new ServicesManagerView(icon, template, services_manager, prev);
		if(service != null){
			/// Label of cancel button
			selector.add_button(new Gtk.Button.with_label(_("Cancel")), Gtk.ResponseType.CANCEL);
		}
		else{
			/// Label of quit button
			selector.add_button(new Gtk.Button.with_label(_("Quit")), Gtk.ResponseType.CLOSE);
		}
		/// Button on service selector screen
		selector.add_button(new Gtk.Button.with_label(_("Use this service")), Gtk.ResponseType.OK);
		selector.response.connect(on_selector_response);
		var event_box = new Gtk.EventBox();
		event_box.add(selector);
		main_view.enable_navigation = false;
		main_window.set_overlay(event_box);
		event_box.show_all();
		update_window_title();
	}
	
	/**
	 * Called when service is selected
	 */
	private void on_selector_response(int response_id){
		var new_service = selector.service;
		destroy_selector();
		switch(response_id){
		case Gtk.ResponseType.OK:
			if(service == null || service.id != new_service.id){
				load_service(new_service);
				ui_state[LAST_SERVICE_KEY] = new_service.id;
			}
			break;
		case Gtk.ResponseType.CLOSE:
			quit();
			break;
		}
	}
	
	/**
	 * Destroys the Service selector screen
	 */
	private void destroy_selector(){
		return_if_fail(selector != null);
		selector.response.disconnect(on_selector_response);
		selector = null;
		
		var overlay = main_window.remove_overlay();
		if(overlay != null){
			overlay.destroy();
		}
		
		main_view.enable_navigation = true;
		update_window_title();
	}
	
	/**
	 * Opens About dialog.
	 */
	private void about(){
		var dialog = new AboutDialog(this);
		dialog.run();
		dialog.destroy();
	}
	
	/**
	 * Opens Preferences dialog with General tab selected.
	 */
	private void edit_settings(){
		show_preferences();
	}
	
	/**
	 * Opens Preferences dialog with Service tab selected.
	 */
	private void edit_service(){
		show_preferences(PreferencesTab.SERVICE);
	}
	
	/**
	 * Opens Preferences dialog with Extensions tab selected.
	 */
	private void edit_extensions(){
		show_preferences(PreferencesTab.EXTENSIONS);
	}
	
	/**
	 * Opens Preferences dialog
	 * 
	 * @param tab	a tab to be initially visible
	 */
	private void show_preferences(PreferencesTab tab=PreferencesTab.GENERAL){
		if(preferences_window != null){
			preferences_window.present();
			return;
		}
		
		preferences_window = new PreferencesWindow(objects, this, this.service, this.storage, this.connection, services_manager, extensions, tab);
		preferences_window.present();
		preferences_window.response.connect(on_preferences_window_response);
	}
	
	private void on_preferences_window_response(int response_id){
		if(preferences_window == null) return;
		if(response_id == Gtk.ResponseType.HELP){
			open_uri_with_fallback(PREFERENCES_HELP_URL);
		}
		else{
			preferences_window.close();
			preferences_window = null;
		}
	}
	
	/**
	 * Opens project support page
	 */
	private void open_support(){
		open_uri_with_fallback(SUPPORT_URL);
	}
	
	/**
	 * Opens release notes for current release
	 */
	private void open_release_notes(){
		open_uri_with_fallback(SERIES_URL);
	}
	
	/**
	 * Called when application is about to be terminated
	 */
	public void on_shutdown(){
		debug("Early shutdown");
		if(player != null){
			player.pause();
		}
		
		if (preferences_window != null){
			preferences_window.close();
			preferences_window = null;
		}
		
		if(main_window != null && main_window.visible){
			this.main_window.hide();
		}
		
		extensions.unload_all();
	}
	
	/**
	 * Called when application is about to be terminated,
	 * but after all early callbacks
	 */
	public void on_late_shutdown(){
		return_if_fail(ui_state != null);
		debug("Late shutdown");
		
		ui_state.set_bool(APP_CRASHED, false);
		try{
			ui_state.save();
			debug("User interface state saved.");
		}
		catch(Error e){
			warning("Unable to save UI state: %s", e.message);
		}
		
		if(main_view != null){
			debug("Unparent main_view");
			var parent = main_view.get_parent() as Gtk.Container;
			if(parent != null){
				parent.remove(main_view);
			}
			debug("Destroy main_view");
			main_view.destroy();
			debug("Destroy main_window");
			main_window.destroy();
			debug("User interface destroyed.");
		}
		
		#if WEB_CACHE
			connection.dump_cache();
		#endif
	}
	
	/**
	 * Called when the main window is about to be closed.
	 * 
	 * @param quit		whether the application should be terminated
	 */
	private void on_close_action(ref bool quit){
		var hide_window = config.hide_window;
		if(quit || hide_window == OptionState.OFF
		|| hide_window == OptionState.AUTO && player.playback_state != STATE_PLAYING
		){
			debug("On close: quit");
			this.quit();
		}
		else{
			debug("On close: hide");
		}
	}
	
	/**
	 * Purge cookies and reload the web interface.
	 */
	private void purge_cookies(){ 
		this.connection.set_up_cookies(true);
		if(main_view != null){
			main_view.reload();
		}
	}
	
	/**
	 * Purge settings and show request to restart application.
	 */
	private void purge_settings(){
		try{
			config.purge();
		}
		catch(Error e){
			critical("Unable to purge configuration: %s", e.message);
		}
		
		var dir = storage.get_child(SERVICES_DIRNAME).user_config_dir;
		try{
			Diorite.System.purge_directory_content(dir);
		}
		catch(GLib.Error e){
			critical("Unable to purge services' conf: %s", e.message);
		}
		
		if(loaded){
			/// Title in information dialog
			var dialog = new Diorite.Widgets.InfoDialog(_("Default settings were restored."),
				/// Request to restart application
				_("Please restart application to finish the process."));
			dialog.run();
		}
	}
	
	/**
	 * Purge user interface state and show request to restart application.
	 */
	private void purge_ui_state(){
		try{
			ui_state.purge();
		}
		catch(Error e){
			critical("Unable to purge UI state: %s", e.message);
		}
		
		if(loaded){
			main_view.go_home();
			/// Title in information dialog
			var dialog = new Diorite.Widgets.InfoDialog(_("Default settings were restored."),
				/// Request to restart application
				_("Please restart application to finish the process."));
			dialog.run();
		}
	}
	
	/**
	 * Disables HTTP proxy
	 */
	private void disable_proxy(){
		try{
			config.proxy_settings = ProxySettings.NONE;
			config.save();
		}
		catch(Error e){
			critical("Unable to save configuration: %s", e.message);
		}
	}

	/**
	 * 
	 */
	private void purge_cache(){
		var dialog = (Gtk.MessageDialog) Object.new(typeof(Gtk.MessageDialog),
				title: "",
				modal: true,
				message_type: Gtk.MessageType.INFO
			);
		/// Action label
		dialog.text = _("Remove files from cache");
		/// Operation in progress
		dialog.secondary_text = _("Operation is in progress. Please wait.");
		dialog.show_all();
		Idle.add(() => {
			purge_cache_cb(dialog);
			return false;
		});
		dialog.run();
	}
	
	/**
	 * Purge cache and reload the web interface.
	 * 
	 * @param dialog	progress dialog to destroy when action is finished.
	 */
	private void purge_cache_cb(Gtk.MessageDialog dialog){
		while(Gtk.events_pending()){
			Gtk.main_iteration();
		}
		#if WEB_CACHE
			connection.clear_cache();
		#endif
		while(Gtk.events_pending()){
			Gtk.main_iteration();
		}
		var dir = storage.get_default_cache_dir();
		try{
			Diorite.System.purge_directory_content(dir, true);
		}
		catch(GLib.Error e){
			critical("Unable to purge album art: %s", e.message);
		}
		while(Gtk.events_pending()){
			Gtk.main_iteration();
		}
		
		if(main_view != null){
			main_view.reload_bypass_cache();
		}
		dialog.destroy();
	}
	
	/**
	 * Called when service is loaded
	 */
	private void on_service_ready(JSApi js_api){
		js_api.load_finished.disconnect(on_service_ready);
		var autoplay = (Args.autoplay || config.auto_play) && !Args.noautoplay;
		debug("Autoplay: %s", autoplay.to_string());
		if(autoplay){
			player.play();
		}
	}
	
	/**
	 * Called when navigation to new page has been approved.
	 */
	private void on_navigation_approved(string uri){
		return_if_fail(service != null);
		ui_state[Service.CONFIG_PREFIX + "." + service.id + ".last_uri"] = uri;
	}
	
	/**
	 * Called when settings have been changed.
	 */
	private void on_settings_changed(string key){
		switch(key){
		case Settings.WEB_PLUGINS:
			main_view.web_plugins = config.web_plugins;
			break;
		#if WEB_CACHE
		case Settings.DATA_CACHE:
			if(config.data_cache){
				connection.enable_cache(config.data_cache_size);
			}
			else{
				connection.disable_cache();
			}
			break;
		case Settings.DATA_CACHE_SIZE:
			connection.enable_cache(config.data_cache_size);
			break;
		#endif
		case Settings.PROXY_SETTINGS:
		case Settings.PROXY_TYPE:
		case Settings.PROXY_SERVER:
		case Settings.PROXY_PORT:
			try{
				connection.set_up_proxy(config.proxy_settings,
					config.proxy_type, config.proxy_server,
					config.proxy_port);
			}
			catch(ProxyError e){
				critical("Unable to set up proxy: %s", e.message);
			}
			break;
		
		}
	}
	
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		update_window_title();
	}
	
	private void update_window_title(){
		if(selector != null){
			/// Title of service selector window
			main_window.title = _("Select service") + " - " + display_name;
		}
		else if(service == null){
			main_window.title = display_name;
		}
		else if(player.playback_state == STATE_NONE || player.song == null){
			main_window.title = "%s - %s".printf(service.name, display_name);
		}
		else if(player.artist == null){
			main_window.title = "%s - %s - %s".printf(player.song, service.name, display_name);
		}
		else{
			main_window.title = "%s - %s - %s - %s".printf(player.song, player.artist, service.name, display_name);
		}
	}
}

} // namespace Nuvola
