/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using JS;
using Diorite.Logger;

namespace Nuvola.JSTools{
/**
 * Converts JavaScriptCore string to Vala UTF-8 string
 *
 * @param jsstring JavaScriptCore string
 * @return Vala UTF-8 string
 */
public static string utf8_string(JS.String jsstring){
	string str = string.nfill(jsstring.get_maximum_utf8_string_size(), ' ');
	// TODO: check real size!
//~ 		var size = jsstring.get_utf8_string(str, str.length);
//~ 		return strsubstring(0, (long) size);
	jsstring.get_utf8_string(str, str.length);
	return str;
}

/**
 * Creates JavaScript object from JSON.
 * 
 * @param ctx	JavaScript context
 * @param json	object in JSON notation
 * @return	JavaScript object
 */
public unowned JS.Value object_from_JSON(JS.Context ctx, string json){
	unowned JS.Value? result = value_from_JSON(ctx, new JS.String(json == "" ? "{}" : json));
	if(result == null || !result.is_object(ctx)){
		result = ctx.make_object();
	}
	return result;
}

/**
 * Creates JavaScript Exception object.
 * 
 * @param ctx	JavaScript context
 * @param message	Error message (single line);
 * @return	JavaScript object
 */
public unowned JS.Value create_exception(JS.Context ctx, string message){
	var exception = """{"type":"NuvolaError", "message":"%s"}""".printf(message.replace("\"", "\\\""));
	debug(exception);
	return object_from_JSON(ctx, exception);
}

/**
 * Obtains string property of JavaScript object
 * 
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @return string value or null if property is not a string
 */
public static string? o_get_string(Context ctx, JS.Object obj, string property){
	unowned JS.Value value = obj.get_property(ctx, new JS.String(property));
	if(value.is_string(ctx)){
		return utf8_string(value.to_jsstring(ctx));
	}
	return null;
}

/**
 * Obtains number property of JavaScript object
 * 
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @return double value or 0.0 if property is not a number
 */
public static double o_get_number(Context ctx, JS.Object obj, string property){
	unowned JS.Value value = obj.get_property(ctx, new JS.String(property));
	if(value.is_number(ctx)){
		return value.to_number(ctx);
	}
	return 0.0;
}

/**
 * Sets string property of JavaScript object
 * 
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @param value property value
 */
public static void o_set_string(Context ctx, JS.Object obj, string property, string value){
	obj.set_property(ctx,
		new JS.String(property),
		JS.Value.string(ctx, new JS.String(value))
	);
}

/**
 * Sets string property of JavaScript object
 * 
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @param value property value
 */
public static void o_set_number(Context ctx, JS.Object obj, string property, double value){
	obj.set_property(ctx,
		new JS.String(property),
		JS.Value.number(ctx, value)
	);
}

/**
 * Obtains object property of JavaScript object
 * 
 * @param ctx JavaScript context
 * @param obj JavaScript object
 * @param property property name
 * @return object or null if property is not a object
 */
public static unowned JS.Object? o_get_object(Context ctx, JS.Object obj, string property){
	unowned JS.Value value = obj.get_property(ctx, new JS.String(property));
	if(value.is_object(ctx)){
		return value.to_object(ctx);
	}
	return null;
}

/**
 * Converts JavaScript value to string
 * 
 * @param ctx JavaScript context
 * @param val JavaScript value
 * @return string value or null if value is not a string
 */
public static string? string_or_null(JS.Context ctx, JS.Value val){
	if(val.is_string(ctx)){
		return utf8_string(val.to_jsstring(ctx));
	}
	return null;
}

public string? value_to_string(JS.Context ctx, JS.Value value){
	if(value.is_string(ctx)){
		return utf8_string(value.to_jsstring(ctx));
	}
	if(value.is_number(ctx)){
		return (value.to_number(ctx).to_string());
	}
	if(value.is_object(ctx)){
		return utf8_string(value_to_JSON(ctx, value.to_object(ctx), 0, null));
	}
	return null;
}

public string? exception_to_string(JS.Context ctx, JS.Value value){
	if(value.is_object(ctx)){
		unowned JS.Object obj = value.to_object(ctx);
		var message = o_get_string(ctx, obj, "message");
		if(message != null){
			var name = o_get_string(ctx, obj, "name");
			var line = (int) o_get_number(ctx, obj, "line");
			var file = o_get_string(ctx, obj, "sourceURL");
			if(line == 0 && file == null){
				return "%s: %s. Enable JS debugging for more details.".printf(name ?? "null", message);
			}
			return "%s:%d: %s: %s".printf(file ?? "(null)", line, name ?? "null", message);
		}
	}
	return value_to_string(ctx, value);
}

#if !HAVE_JSCORE_JSON
// WebKitGtk < 1.3 does not have JSValueMakeFromJSONString
/**
 * Deserializes value from JSON notation.
 * 
 * @param ctx		JavaScript context
 * @param js_string	string in JSON format
 * @return JavaScript value
 */
public unowned JS.Value value_from_JSON (Context ctx, String js_string){
	JS.Value? exception = null;
	debug("Custom implementation of value_from_JSON.");
	unowned JS.Object window = ctx.get_global_object();
	unowned JS.Object json = o_get_object(ctx, window, "JSON");
	unowned JS.Object parse = o_get_object(ctx, json, "parse");
	
	JS.Value[] args = {null};
	Diorite.Misc.array_add_unowned((Diorite.Misc.UnownedItem[])args, 0, (Diorite.Misc.UnownedItem) JS.Value.string(ctx, js_string));
	
	unowned JS.Value result = parse.call_as_function(ctx, json, (JS.Value[]) args, out exception);
	debug("Emd of custom implementation of value_from_JSON.");
	if(exception == null){
		return result;
	}
	else{
		return JS.Value.null(ctx);
	}
}

// WebKitGtk < 1.3 does not have JSValueCreateJSONString
/**
 * Serializes value to JSON notation.
 * 
 * @param ctx			JavaScript context
 * @param value		value to serialize
 * @param indent		indentation
 * @param exception	JavaScript exception
 * @return	JavaScript string with serialized value
 */
public JS.String value_to_JSON (Context ctx, JS.Value value, uint indent, out JS.Value? exception=null){
	exception = null;
	debug("Custom implementation of value_to_JSON.");
	unowned JS.Object window = ctx.get_global_object();
	unowned JS.Object json = o_get_object(ctx, window, "JSON");
	unowned JS.Object stringify = o_get_object(ctx, json, "stringify");
	
	JS.Value[] args = {null};
	Diorite.Misc.array_add_unowned((Diorite.Misc.UnownedItem[])args, 0, (Diorite.Misc.UnownedItem) value);			
	
	unowned JS.Value? result = stringify.call_as_function(ctx, json, (JS.Value[]) args, out exception);
	debug("Emd of custom implementation of value_to_JSON.");
	if(exception != null || result == null){
		return new JS.String("");
	}
	if(result.is_string(ctx)){
		return result.to_jsstring(ctx);
	}
	else{
		return new JS.String("");
	}
}
#else
/**
 * Serializes value to JSON notation.
 * 
 * @param ctx			JavaScript context
 * @param value		value to serialize
 * @param indent		indentation
 * @param exception	JavaScript exception
 * @return	JavaScript string with serialized value
 */
public JS.String value_to_JSON (Context ctx, JS.Value value, uint indent, out JS.Value? exception=null){
	return value.to_JSON(ctx, indent, out exception);
}

/**
 * Deserializes value from JSON notation.
 * 
 * @param ctx		JavaScript context
 * @param js_string	string in JSON format
 * @return JavaScript value
 */
public unowned JS.Value value_from_JSON (Context ctx, String js_string){
	return JS.Value.from_JSON(ctx, js_string);
}
#endif

} // namespace Nuvola.JSTools
