/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using WebKit;
using Soup;
using Gee;
using Diorite;
using Diorite.Logger;
using Diorite.Widgets;
using Diorite.Tools;

namespace Nuvola{

const string STATE_PLAYING = "playing";
const string STATE_PAUSED = "paused";
const string STATE_NONE = "none";


namespace Actions{
	const string PLAY = "play";
	const string PAUSE = "pause";
	const string TOGGLE_PLAY = "toggle-play";
	const string PREV_SONG = "prev";
	const string NEXT_SONG = "next";
	const string THUMBS_UP = "thumbs-up";
	const string THUMBS_DOWN = "thumbs-down";
	const string FAVORITE = "favorite";
}

const string PLAYER_ACTIONS_EXTRA[3] = {Actions.THUMBS_UP, Actions.THUMBS_DOWN, Actions.FAVORITE};
const string PLAYER_ACTIONS[3] = {Actions.PREV_SONG, Actions.TOGGLE_PLAY, Actions.NEXT_SONG};
const string PLAYER_ACTIONS_ALL[6] = {Actions.PREV_SONG, Actions.TOGGLE_PLAY, Actions.NEXT_SONG, Actions.THUMBS_UP, Actions.THUMBS_DOWN, Actions.FAVORITE};

public class Player : GLib.Object{
	/// Action label
	static string PLAY_LABEL = _("Play");
	/// Action label
	static string PAUSE_LABEL = _("Pause");
	static string PLAY_ICON = "media-playback-start";
	static string PAUSE_ICON = "media-playback-pause";
	private JSApi? _js_api = null;
	public JSApi? js_api{
		get{
			return _js_api;
		}
		set{
			if(_js_api != null){
				_js_api.song_changed.disconnect(on_song_changed);
				_js_api.load_finished.disconnect(on_load_finished);
				if(playback_state == STATE_PLAYING){
					pause();
				}
			}
			
			_js_api = value;
			song_changed(null, null, null, null);
			playback_state = STATE_NONE;
			actions.get_action(Actions.PREV_SONG).sensitive = false;
			actions.get_action(Actions.NEXT_SONG).sensitive = false;
			actions.get_action(Actions.THUMBS_DOWN).sensitive = false;
			actions.get_action(Actions.THUMBS_UP).sensitive = false;
			actions.get_action(Actions.FAVORITE).sensitive = false;
			
			if(_js_api != null){
				_js_api.song_changed.connect(on_song_changed);
				_js_api.load_finished.connect(on_load_finished);
			}
		}
	}
	
	public Service? service{get; set; default = null;}
	
	private weak Diorite.Application app;
	private weak Configuration config;
	private Diorite.Storage storage;
	private Diorite.Connection connection;
	private File? album_art_file = null;
	private weak Diorite.Actions? actions;
	private weak Diorite.Action action_toggle;
	private weak Diorite.Action action_play;
	private weak Diorite.Action action_pause;
	private int8 album_art_number = -1;
	
	[Description(nick = "playback state", blurb = "playback state: 'playing', 'paused' or 'none'")]
	public string? playback_state {get; private set; default = STATE_NONE;} 
	
	[Description(nick = "URL of album art", blurb = "URL of album art")]
	private string? album_art_url {get; private set;} 
	
	[Description(nick = "Path to album art", blurb = "Path to album art")]
	public string? album_art {get; private set;} 
	
	[Description(nick = "Path to album art", blurb = "Path to album art")]
	public string? album {get; private set;} 
	
	[Description(nick = "name of the current artist", blurb = "name of the current artist")]
	public string? artist {get; private set;} 
	
	[Description(nick = "name of the current song", blurb = "name of the current song")]
	public string? song {get; private set;} 
	
	public Player(Diorite.Application app, Configuration config, Storage storage, Diorite.Connection connection, Diorite.Actions? actions){
		this.app = app;
		this.config = config;
		this.actions = actions;
		this.connection = connection;
		this.storage = storage;
		append_ui_actions();
	}
	
	/**
	 * Response to frame events - updates current status, stores album art
	 * and emit associated signals
	 */
	private void on_song_changed(string? song, string? artist, string? album, string? art_uri, string playback_state){
		
		if(this.playback_state != playback_state){
			debug("Playback state changed: %s → %s",
				this.playback_state ?? "null", playback_state ?? "null");
			this.playback_state = playback_state;
			switch(playback_state){
 			case STATE_PAUSED:
				action_toggle.stock_id = PLAY_ICON;
				action_toggle.icon_name = PLAY_ICON;
				action_toggle.label = PLAY_LABEL;
				action_toggle.sensitive = true;
				action_play.sensitive = true;
				action_pause.sensitive = false;
				break;
			case STATE_PLAYING:
				action_toggle.stock_id = PAUSE_ICON;
				action_toggle.icon_name = PAUSE_ICON;
				action_toggle.label = PAUSE_LABEL;
				action_toggle.sensitive = true;
				action_play.sensitive = false;
				action_pause.sensitive = true;
				break;
			default:
				action_toggle.stock_id = PLAY_ICON;
				action_toggle.icon_name = PLAY_ICON;
				action_toggle.label = PLAY_LABEL;
				action_toggle.sensitive = false;
				action_play.sensitive = false;
				action_pause.sensitive = false;
				break;
			}
		}
		
		var song_changed = false;
		if(this.album_art_url != art_uri){
			debug("Album art changed: %s → %s",
				this.album_art_url ?? "null", art_uri ?? "null");
			this.album_art_url = art_uri;
			if(album_art_file != null){
				try{
					album_art_file.delete();
				}
				catch(Error e){
					warning("Unable to delete old album art file: %s", e.message);
				}
			}
			
			/* Some clients don't reload album art if it has same name, so we reuse file names
			 * album_art.0 - album_art.9. Album art file is not deleted, when Nuvola Player exits,
			 * but the file will be reused in the next run. */
			album_art_number = album_art_number >= 9 ? 0 : album_art_number + 1;
			album_art_file = storage.get_default_cache_file("album_art." + album_art_number.to_string());
			
			if(art_uri != null && connection.download_file(art_uri, album_art_file)){
				this.album_art = album_art_file.get_path();
			}
			else{
				album_art_file = null;
				var icon = Gtk.IconTheme.get_default().lookup_icon(app.icon, 128, 0);
				if(icon != null){
					this.album_art = icon.get_filename();
					/* If Nuvola Palyer is run from source directory via wrapper, IconTheme returns
					 * relative path to icon instead of absolute. */
					if(this.album_art.has_prefix("./")){
						this.album_art = Environment.get_current_dir() + this.album_art.substring(1);
					}
				}
				else{
					this.album_art = null;
				}
			}
			song_changed = true;
		}
		
		if(this.artist != artist){
			debug("Artist changed: %s → %s",
				this.artist ?? "null", artist ?? "null");
			this.artist = artist;
			song_changed = true;
		}
		
		if(this.song != song){
			debug("Song changed: %s → %s",
				this.song ?? "null", song ?? "null");
			this.song = song;
			song_changed = true;
		}
		
		if(this.album != album){
			debug("Album changed: %s → %s",
				this.album ?? "null", album ?? "null");
			this.album = album;
			song_changed = true;
		}
		
		if(song_changed){
			this.song_changed(song, artist, album, album_art);
		}
	}
	
	private void on_load_finished(){
		if(service == null) return;
		try{
			js_api.execute_script_from_file(service.integration_file);
		}
		catch(JSError e){
			warning("Unable to load service integration file: %s", e.message);
		}
	}
	
	/**
	 * Emitted when song is changed.
	 * 
	 * @param song name of the current song
	 * @param artist name of the current artist
	 * @param album album name
	 * @param album_art album art file
	 *  
	 */
	public signal void song_changed(string? song, string? artist, string? album, string? album_art);
	
	/**
	 * Trigger player action
	 * 
	 * @param action action
	 * @param notify if true a notification will be shown if needed
	 */
	public void trigger_action(string action){
		return_if_fail(js_api != null);
		try{
			this.js_api.send_message(action);
		}
		catch(JSError e){
			critical("Player: %s", e.message);
		}
	}
	
	
	/**
	 * Skips to the next track in the tracklist.
	 * 
	 * If there is no next track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void next_song(){
		this.trigger_action(Actions.NEXT_SONG);
	}
	
	/***
	 * Skips to the previous track in the tracklist.
	 * 
	 * If there is no previous track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void previous_song(){
		this.trigger_action(Actions.PREV_SONG);
	}
	
	/**
	 * Pauses playback.
	 * 
	 * If playback is already paused, this has no effect. Calling Play after this should
	 * cause playback to start again from the same position.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void pause(){
		this.trigger_action(Actions.PAUSE);
	}
	
	/**
	 * Pauses playback.
	 * If playback is already paused, resumes playback.
	 * If playback is stopped, starts playback.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void toggle_play(){
		this.trigger_action(Actions.TOGGLE_PLAY);
	}
	
	/**
	 * Stops playback.
	 * 
	 * If playback is already stopped, this has no effect. Calling Play after this should
	 * cause playback to start again from the beginning of the track. 
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void stop(){
		trigger_action(Actions.PAUSE);
	}
	
	/**
	 * Starts or resumes playback.
	 * 
	 * If already playing, this has no effect.
	 * If there is no track to play, this has no effect.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void play(){
		this.trigger_action(Actions.PLAY);
	}
	
	/**
	 * Sets Thumbs Up (= awesome) to song
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void thumbs_up(){
		this.trigger_action(Actions.THUMBS_UP);
	}
	
	/**
	 * Sets Thumbs Dowb (= horrible) to song
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void thumbs_down(){
		this.trigger_action(Actions.THUMBS_DOWN);
	}
	
	/**
	 * Marks song as favorited
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void favorite(){
		this.trigger_action(Actions.FAVORITE);
	}
	
	/**
	 * Creates UI actions and appends them to UIActions instance
	 */
	private void append_ui_actions(){
		if(actions == null) return;
		
		/* Normal items. Name, stock id, label, accelerator, tooltip, callback */
		
		actions.add_actions({
			/// Action tooltip
			new NormalAction(Actions.TOGGLE_PLAY, PLAY_ICON, PLAY_LABEL, get_pp_accel(), _("Toggle play/pause"), toggle_play),
			new NormalAction(Actions.PLAY, PLAY_ICON, PLAY_LABEL, null, PLAY_LABEL, play),
			/// Tool bar item tooltip
			new NormalAction(Actions.PAUSE, PAUSE_ICON, PAUSE_LABEL, null, PAUSE_LABEL, pause),
			/// Action label
			new NormalAction(Actions.PREV_SONG, "media-skip-backward", _("Previous song"), 
			/// Action tooltip
			"<alt>Left", _("Skip to previous song"), previous_song),
			/// Action label
			new NormalAction(Actions.NEXT_SONG, "media-skip-forward", _("Next song"), "<alt>Right", 
			/// Action tooltip
			_("Skip to next song"), next_song),
			/// Short label for command 'thumb up' or 'like this song' (see Google Music buttons)
			new NormalAction(Actions.THUMBS_UP, null, _("Thumbs Up"), null, null, thumbs_up),
			/// Short label for command 'thumb down' or 'dislike this song' (see Google Music buttons)
			new NormalAction(Actions.THUMBS_DOWN, null, _("Thumbs Down"), null, null, thumbs_down),
			/// Short label for command 'add this song to favorites'
			new NormalAction(Actions.FAVORITE, null, _("Favorite"), null, null, favorite)
		});
		
		action_toggle = actions.get_action(Actions.TOGGLE_PLAY);
		action_toggle.sensitive = false;
		actions.get_action(Actions.PREV_SONG).sensitive = false;
		actions.get_action(Actions.NEXT_SONG).sensitive = false;
		action_play = actions.get_action(Actions.PLAY);
		action_play.sensitive = false;
		action_pause = actions.get_action(Actions.PAUSE);
		action_pause.sensitive = false;
		
		
		config.notify["space-key"].connect_after(() => {
			action_toggle.keybinding = get_pp_accel();
		});
	}
	
	private string get_pp_accel(){
		return config.space_key ? "<ctrl>space" : "<ctrl>P";
	}
}

} // namespace Nuvola
