/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using WebKit;
using Soup;
using Gee;
using Diorite;
using Diorite.Logger;
using Diorite.Widgets;
using Diorite.Tools;
using Diorite.System;
using JS;


/**
 * Service integration class
 */
public class Nuvola.Service: GLib.Object{
	/**
	 * Prefix used for service configuration options.
	 */
	public static const string CONFIG_PREFIX = "service";
	/**
	 * Name of file with integration script.
	 */
	private static const string INTEGRATION_FILENAME = "integration.js";
	/**
	 * Name of file with service description in HTML.
	 */
	private static const string DESCRIPTION_FILENAME = "description.html";
	
	[Description(nick = "service identifier", blurb = "Unique identitier of service")]
	public string id{
		get;
		private set;
	}
	
	[Description(nick = "name", blurb = "Name of the service")]
	public string name{
		get;
		private set;
	}
	
	[Description(nick = "major version", blurb = "Major version of service integration")]
	public int version{
		get;
		private set;
	}
	
	[Description(nick = "minor version", blurb = "Minor version of service integration")]
	public int version_minor{
		get;
		private set;
	}
	
	[Description(nick = "maintainer name", blurb = "Name of the maintainer")]
	public string maintainer_name{
		get;
		private set;
	}
	
	[Description(nick = "maintainer link", blurb = "Link to the maintainer")]
	public string maintainer_link{
		get;
		private set;
	}
	
	[Description(nick = "Resources URI", blurb = "URI od directory with service resources")]
	public File resources{
		get;
		private set;
	}
	
	[Description(nick = "Major API version", blurb = "Minimal required major API version")]
	public int api_major{
		get;
		private set;
	}
	
	[Description(nick = "Minor API version", blurb = "Minimal required minor API version")]
	public int api_minor{
		get;
		private set;
	}
	
	private File config;
	/**
	 * main page of the web interface for music playback
	 */
	[Description(nick = "Home page", blurb = "Main page of the web interface")]
	public string home_page{
		get;
		private set;
	}
	
	[Description(nick = "Flash plugin", blurb = "Whether Flash plugin has to be enabled")]
	public bool flash_plugin{
		get;
		private set;
	}
	
	[Description(nick = "Removable", blurb = "Whether the service can be removed")]
	public bool removable{
		get;
		private set;
	}
	/**
	 * pages not matching this regular expression will be opened in the default web browser
	 */
	private Regex sandbox;
	
	public File integration_file{ get; private set; }
	private File description_file;
	
	private string? _description = null;
	[Description(nick = "description", blurb = "Description of the service in HTML format.")]
	public string description{
		get{
			if(_description == null){
				try{
					_description = read_file(description_file);
				}
				catch(Error e){
					warning("Unable to read description: %s", e.message);
					/// Shown in Service selector if service doesn't provide description
					return _("No description provided for this service.");
				}
			}
			
			return _description;
		}
	}
	
	public File settings_script{
		owned get{
			return resources.get_child("settings.js");
		}
	}
	
	/**
	 * Loads new service integration
	 * 
	 * @param id			service id
	 * @param version		version of service integration
	 * @param integration	integration script
	 * @param sandbox		sandbox expression
	 */
	public Service(string id, MultiTypeMap metadata, File config, File resources, bool removable=false) throws ServiceError{
		this.id = id;
		this.config = config;
		this.integration_file = resources.get_child(INTEGRATION_FILENAME);
		this.removable = removable;
		
		if(integration_file.query_file_type(0) != FileType.REGULAR){
			throw new ServiceError.LOADING_FAILED(
				"Integration file not found.");
		}
		
		description_file = resources.get_child(DESCRIPTION_FILENAME);
		if(description_file.query_file_type(0) != FileType.REGULAR){
			throw new ServiceError.LOADING_FAILED(
				"Description file not found.");
		}
		
		this.resources = resources;
		
		name = metadata["name"] ?? "";
		if(name == ""){
			throw new ServiceError.LOADING_FAILED(
				"Empty 'name' entry");
		}
		
		var sandbox_pattern = metadata["sandbox_pattern"] ?? "";
		if(sandbox_pattern == ""){
			throw new ServiceError.LOADING_FAILED(
				"Empty 'sandbox_pattern' entry");
		}
		
		try{
			sandbox = new Regex(sandbox_pattern,
					RegexCompileFlags.CASELESS | RegexCompileFlags.ANCHORED);
		}
		catch(RegexError e){
			throw new ServiceError.LOADING_FAILED(
				"Invalid 'sandbox_pattern' entry '%s': '%s'"
				.printf(sandbox_pattern, e.message));
		}
		
		maintainer_name = metadata["maintainer_name"] ?? "";
		if(maintainer_name == ""){
			throw new ServiceError.LOADING_FAILED(
				"Empty 'maintainer_name' entry");
		}
		
		home_page = metadata["home_page"] ?? "";
		if(home_page.has_prefix("service://")){
			home_page = resources.get_child(home_page.substring(10)).get_uri();
		}
		else if(!home_page.has_prefix("http://")
		&&  !home_page.has_prefix("https://")){
			throw new ServiceError.LOADING_FAILED(
				"Empty or invalid 'home_page' entry '%s'".printf(home_page));
		}
		
		maintainer_link = metadata["maintainer_link"] ?? "";
		if(!maintainer_link.has_prefix("http://")
		&&  !maintainer_link.has_prefix("https://")
		&&  !maintainer_link.has_prefix("mailto:")){
			throw new ServiceError.LOADING_FAILED(
				"Empty or invalid 'maintainer_link' entry: '%s'"
				.printf(maintainer_link));
		}
		
		version = metadata.get_int("version", 0);
		if(version <= 0){
			throw new ServiceError.LOADING_FAILED(
				"Invalid 'version' entry: '%d'".printf(version));
		}
		
		version_minor = metadata.get_int("version_minor", 0);
		if(version_minor < 0){
			throw new ServiceError.LOADING_FAILED(
				"Invalid 'version_minor' entry: '%d'".printf(version_minor));
		}
		
		api_major = metadata.get_int("api_major", 1);
		if(api_major <= 0){
			throw new ServiceError.LOADING_FAILED(
				"Invalid 'api_major' entry: '%d'".printf(api_major));
		}
		
		api_minor = metadata.get_int("api_minor", 0);
		if(api_minor < 0){
			throw new ServiceError.LOADING_FAILED(
				"Invalid 'api_minor' entry: '%d'".printf(api_minor));
		}
		
		if(!JSApi.is_supported(api_major, api_minor)){
			throw new ServiceError.LOADING_FAILED(
				"Requested unsupported api: %d.%d'".printf(api_major, api_minor));
		}
		flash_plugin = metadata.get_bool("flash_plugin", true);
	}
	
	/**
	 * Decide whether given URI is allowed in sandbox.
	 * 
	 * @param uri	URI to check
	 * @return		true if UIR matches sandbox
	 */
	public bool matches_sandbox(string uri){
		return sandbox.match(uri);
	}
	
	public string config_json{
		owned get{
			try{
				return read_file(config);
			}
			catch(Error e){
				return "";
			}
		}
		set{
			try{
				overwrite_file(config, value);
			}
			catch(Error e){
				warning("Unable to save srvice configuration: %s", e.message);
			}
		}
	}
	
	#if DEBUG_MEMORY
	~Service(){
		debug("~Service");
	}
	#endif
}
