/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 using Diorite.Logger;

namespace Nuvola{

private const string SERVICES_DIRNAME = "services";
	
/**
 * Services Manager deals with management and loading of service integrations.
 */
public class ServicesManager: GLib.Object{
	private Diorite.Storage storage;
	/**
	 * Name of file with metadata.
	 */
	private static const string METADATA_FILENAME = "metadata.conf";
	
	/**
	 * Regular expression to check validity of service identifier
	 */
	private static Regex id_regex;
	
	/**
	 * Whether compatible Flash plugin is available
	 */
	public bool compatible_flash_plugin = false;
	
	/**
	 * Creates new service manager
	 * 
	 * @param storage storage with service integrations
	 */
	public ServicesManager(Diorite.Storage storage){
		this.storage = storage;
	}
	
	/**
	 * Loads service by id.
	 * 
	 * @param id service id
	 * @return service
	 */
	public Service? get_service(string id){
		if(!check_id(id)){
			warning("Service id '%s' is invalid.", id);
			return null;
		}
		Service? service = null;
		Service? item;
		
		var service_storage = storage.get_child(id);
		
		var user_dir = service_storage.user_data_dir;
		if(user_dir != null){
			try{
				service = load_service_from_dir(user_dir, true);
				debug("Found service %s at %s, version %d.%d",
					service.name, user_dir.get_path(), service.version, service.version_minor);
			}
			catch(ServiceError e){
				warning("Unable to load service from %s: %s", user_dir.get_path(), e.message);
			}
		}
		
		foreach(var dir in service_storage.data_dirs){
			try{
				item = load_service_from_dir(dir);
				debug("Found service %s at %s, version %d.%d",
					item.name, dir.get_path(), item.version, item.version_minor);
				if(service == null || item.version > service.version
				|| item.version == service.version && item.version_minor > service.version_minor){
					service = item;
				}
			}
			catch(ServiceError e){
				warning("Unable to load service from %s: %s", dir.get_path(), e.message);
			}
		}
		
		if(service != null){
			message("Using service %s from %s, version %d.%d",
				service.name, service.resources.get_path(), service.version, service.version_minor);
		}
		
		return service;
	}
	
	public Service load_service_from_dir(File dir, bool removable=false) throws ServiceError{
		if(dir.query_file_type(0) != FileType.DIRECTORY){
			throw new ServiceError.LOADING_FAILED(@"$(dir.get_path()) is not a directory");
		}
		var id = dir.get_basename();
		var metadata_file = dir.get_child(METADATA_FILENAME);
		if(metadata_file.query_file_type(0) != FileType.REGULAR){
			throw new ServiceError.LOADING_FAILED(@"$(metadata_file.get_path()) is not a file");
		}
		var metadata = new Diorite.MapFile(metadata_file, true);
		return new Service(id, metadata, storage.get_default_config_file(id + ".json"), dir, removable);
	}
	/**
	 * Lists available services
	 * 
	 * @return hash table of service id - metadata pairs
	 */
	public HashTable<string, Service> list_services(){
		HashTable<string,  Service> result =
			new HashTable<string,  Service>(str_hash, str_equal);
		FileInfo file_info;
		Service? service;
		Service? tmp_service;
		var user_dir = storage.user_data_dir;
		if(user_dir != null){
			try{
				var enumerator = user_dir.enumerate_children(
					FILE_ATTRIBUTE_STANDARD_NAME, 0);
				while((file_info = enumerator.next_file()) != null){
					string name = file_info.get_name();
					if(!check_id(name)) continue;
					
					var service_dir = user_dir.get_child(name);
					if(service_dir.query_file_type(0) != FileType.DIRECTORY){
						continue;
					}
					
					try{
						service = load_service_from_dir(service_dir, true);
						debug("Found service %s at %s, version %d.%d",
						service.name, service_dir.get_path(), service.version, service.version_minor);
						result.insert(name, service);
					}
					catch(ServiceError e){
						warning("Unable to load service from %s: %s", service_dir.get_path(), e.message);
					}
				}
			}
			catch(Error e){
				warning("Filesystem error: %s", e.message);
			}
		}
		
		foreach(var dir in storage.data_dirs){
			try{
				var enumerator = dir.enumerate_children(
					FILE_ATTRIBUTE_STANDARD_NAME, 0);
				while((file_info = enumerator.next_file()) != null){
					string name = file_info.get_name();
					if(!check_id(name)) continue;
					
					var service_dir = dir.get_child(name);
					if(service_dir.query_file_type(0) != FileType.DIRECTORY){
						continue;
					}
					
					try{
						service = load_service_from_dir(service_dir);
					}
					catch(ServiceError e){
						warning("Unable to load service from %s: %s", service_dir.get_path(), e.message);
						continue;
					}
					
					debug("Found service %s at %s, version %d.%d",
						service.name, service_dir.get_path(), service.version, service.version_minor);
					
					tmp_service = result.lookup(name);
					
					// Insert new value, if service has not been added yet,
					// or override previous service integration, if
					// the new one has greater version.
					if(tmp_service == null
					|| service.version > tmp_service.version 
					|| service.version == tmp_service.version && service.version_minor > tmp_service.version_minor){
						result.insert(name, service);
					}
				}
			}
			catch(Error e){
				warning("Filesystem error: %s", e.message);
			}
		}
		
		return result;
			
	}
	
	/**
	 * Installs service from package
	 * 
	 * @param package		a package to install from.
	 * @throw ServiceError on failure
	 */
	public Service install_service(File package) throws ServiceError{
		debug("Install service: %s", package.get_path());
		File tmp_dir;
		try{
			tmp_dir = File.new_for_path(DirUtils.make_tmp("nuvolaplayerXXXXXX"));
		}
		catch(FileError e){
			throw new ServiceError.IOERROR(e.message);
		}
		
		string spawn_stdout = null;
		string spawn_stderr = null;
		int result = 0;
		
		try{
			string[] args = {"tar", "-xvzf", package.get_path()};
			Process.spawn_sync(tmp_dir.get_path(), args, null,  SpawnFlags.SEARCH_PATH, null, out spawn_stdout, out spawn_stderr, out result);
		}
		catch(SpawnError e){
			Diorite.System.try_purge_dir(tmp_dir);
			throw new ServiceError.INVALID_FILE(e.message);
		}
		
		debug("result: %d", result);
		debug("out: %s", spawn_stdout);
		debug("err: %s", spawn_stderr);
		if(result != 0){
			if(spawn_stderr != null){
				Diorite.System.try_purge_dir(tmp_dir);
				throw new ServiceError.INVALID_FILE(spawn_stderr);
			}
		}
		
		var control = new Diorite.MapFile(tmp_dir.get_child("control"), true);
		var format = control.get_int("format", 1, 1);
		var service_id = control["service_id"];
		if(format != 2 || service_id == null || service_id == ""){
			Diorite.System.try_purge_dir(tmp_dir);
			throw new ServiceError.INVALID_FILE("Package has wrong format.");
		}
		
		var service_dir = tmp_dir.get_child(service_id);
		try{
			load_service_from_dir(service_dir);
		}
		catch(ServiceError e){
			Diorite.System.try_purge_dir(tmp_dir);
			throw e;
		}
		
		var destination = storage.get_default_data_dir(service_id);
		if(destination.query_exists()){
			try{
				Diorite.System.purge_directory_content(destination, true);
				destination.delete();
			}
			catch(Error e){
				Diorite.System.try_purge_dir(tmp_dir);
				throw new ServiceError.IOERROR(e.message);
			}
		}
		else{
			try{
				destination.get_parent().make_directory_with_parents();
			}
			catch(Error e){
				// Not fatal
			}
		}
		
		result = 0;
		try{
			string[] args = {"mv", "-v", service_dir.get_path(), destination.get_parent().get_path()};
			Process.spawn_sync(tmp_dir.get_path(), args, null,  SpawnFlags.SEARCH_PATH, null, out spawn_stdout, out spawn_stderr, out result);
		}
		catch(SpawnError e){
			Diorite.System.try_purge_dir(tmp_dir);
			throw new ServiceError.IOERROR(e.message);
		}
		debug("result: %d", result);
		debug("out: %s", spawn_stdout);
		debug("err: %s", spawn_stderr);
		if(result != 0){
			if(spawn_stderr != null){
				Diorite.System.try_purge_dir(tmp_dir);
				throw new ServiceError.IOERROR(spawn_stderr);
			}
		}
		
		Diorite.System.try_purge_dir(tmp_dir);
		
		return load_service_from_dir(destination); // throws ServiceError
	}
	
	/**
	 * Removes service.
	 * 
	 * @param service		service to remove
	 * @throw ServiceError on failure
	 */
	public void remove_service(Service service) throws ServiceError{
		var dir = service.resources;
		if(dir == null){
			throw new ServiceError.IOERROR("Invalid service directory");
		}
		if(dir.query_exists()){
			try{
				Diorite.System.purge_directory_content(dir, true);
				dir.delete();
			}
			catch(Error e){
				throw new ServiceError.IOERROR(e.message);
			}
		}
		else{
			throw new ServiceError.IOERROR(@"$(dir.get_path()) does not exist.");
		}
	}
	
	/**
	 * Check if the service identifier is valid
	 * 
	 * @param id service identifier
	 * @return true if id is valid
	 */
	public static bool check_id(string id){
		if(id_regex == null){
			try{
				id_regex = new Regex("^\\w+$");
			}
			catch(RegexError e){
				error("Unable to compile regular expression /^\\w+$/.");	
			}
		}
		return id_regex.match(id);
	}
}

public errordomain ServiceError{
	LOADING_FAILED,
	COMMAND_FAILED,
	INVALID_FILE,
	IOERROR;
}

} // namespace Nuvola
