/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Nuvola.Settings;
using Diorite;

namespace Nuvola{

namespace Settings{
	// Configuration keys
	private const string PROXY_TYPE = "proxy_type";
	private const string PROXY_SERVER = "proxy_server";
	private const string PROXY_PORT = "proxy_port";
	private const string PROXY_SETTINGS = "proxy_settings";
	private const string HIDE_WINDOW = "hide_window";
	private const string PAGE_CACHE = "page_cache";
	private const string DATA_CACHE = "data_cache";
	private const string DATA_CACHE_SIZE = "data_cache_size";
	private const string AUTO_PLAY = "auto_play";
	private const string WEB_PLUGINS = "web_plugins";
	private const string SPACE_KEY = "space_key";
}


/**
 * Interface providing access to configuration options via GObject
 * properties with right type and right default value.
 */
public interface Configuration: MultiTypeMap, Serializable{
	[Description(nick = "Proxy settings", blurb = "Defines what proxy settings is used: no, system or manual settings.")]
	public abstract ProxySettings proxy_settings{ get; set; }
	[Description(nick = "Proxy type", blurb = "Defines what proxy type is used: HTTP(S) or SOCKS 4/5")]
	public abstract ProxyType proxy_type{ get; set; }
	[Description(nick = "Proxy server", blurb = "Address of proxy server.")]
	public abstract string proxy_server{ owned get; set; }
	[Description(nick = "Proxy server port", blurb = "Port of proxy server.")]
	public abstract string proxy_port{ owned get; set; }
	[Description(nick = "Window hiding", blurb = "Defines whether hide window when close button is pressed.")]
	public abstract OptionState hide_window{ get; set; }
	[Description(nick = "Auto play", blurb = "Whether automatically start playback (if possible).")]
	public abstract bool auto_play{ get; set; }
	[Description(nick = "Web plugins", blurb = "Whether web plugins are enabled.")]
	public abstract bool web_plugins{ get; set; }
	[Description(nick = "Size of data cache", blurb = "Defines how much data cache in MB is used to cache remote objects.")]
	public abstract int data_cache_size{ get; set; }
	[Description(nick = "Data cache", blurb = "Enables or disables cache for remote data (images, scripts, stylesheets, etc.).")]
	public abstract bool data_cache{ get; set; }
	[Description(nick = "Use space key", blurb = "Whether it is allowed to use space key in keybindings.")]
	public abstract bool space_key{ get; set; }
}

/**
 * Implementation of Configuration interface using configuration file as storage.
 */
public class ConfigurationFile: MapFile, Configuration{
	/**
	 * Creates new configuration storage using specified file
	 * 
	 * @param file file used to store configuration
	 */
	public ConfigurationFile(File file){
		base(file, false);
	}
	
	public ProxySettings proxy_settings{
		get{
			var val = get_int(PROXY_SETTINGS, ProxySettings.AUTO,
				ProxySettings.AUTO, ProxySettings.NONE);
			return (ProxySettings) val;
		}
		
		set{
			set_int(PROXY_SETTINGS, (int) value);
		}
	}
	
	public ProxyType proxy_type{
		get{
			var val = get_int(PROXY_TYPE, ProxyType.HTTP,
				ProxyType.HTTP, ProxyType.SOCKS);
			return (ProxyType) val;
		}
		
		set{
			set_int(PROXY_TYPE, (int) value);
		}
	}
	
	public string proxy_server{
		owned get{
			return this[PROXY_SERVER] ?? "";
		}
		
		set{
			this[PROXY_SERVER] = value ?? "";
		}
	}
	
	public string proxy_port{
		owned get{
			return this[PROXY_PORT] ?? "";
		}
		
		set{
			this[PROXY_PORT] = value ?? "";
		}
	}
	
	public OptionState hide_window{
		get{
			return get_state(HIDE_WINDOW);
		}
		
		set{
			set_state(HIDE_WINDOW, value);
		}
	}
	
	public bool auto_play{
		get{
			return get_bool(AUTO_PLAY, false);
		}
		set{
			set_bool(AUTO_PLAY, value);
		}
	}
	
	public bool web_plugins{
		get{
			return get_bool(WEB_PLUGINS, false);
		}
		set{
			set_bool(WEB_PLUGINS, value);
		}
	}
	
	public int data_cache_size{
		get{
			return get_int(DATA_CACHE_SIZE, 100, 0, 1024);;
		}
		
		set{
			set_int(DATA_CACHE_SIZE, value);
		}
	}
	
	public bool data_cache{
		get{
			return this.get_bool(DATA_CACHE, true);
		}
		set{
			this.set_bool(DATA_CACHE, value);
		}
	}
	
	public bool space_key{
		get{
			return get_bool(SPACE_KEY, true);
		}
		
		set{
			set_bool(SPACE_KEY, value);
		}
	}
	
} // namespace Nuvola
}
