/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


namespace Nuvola.Extensions.Debugging{

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of an extension
		_("Debugging"),
		Nuvola.VERSION,
		/// Description of extension
		_("This plugin is used for debugging and appends 'Flash Support' and 'Audio Support' tests to menu."),
		"Jiří Janoušek",
		typeof(Extension)
	};
}


/**
 * Simple sample plugin
 */
public class Extension:  Nuvola.Extension{
	private Diorite.Actions actions;
	private uint ui_merge_id = 0;
	private Diorite.Storage storage;
	private Gtk.Window? audio_test = null;
	private Gtk.Window? flash_test = null;
	private weak MainWindow? main_window = null;
	
	~Extension(){
		if(audio_test != null){
			audio_test.destroy();
			audio_test = null;
		}
		if(flash_test != null){
			flash_test.destroy();
			flash_test = null;
		}
	}
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		debug("[%s] load", id);
		// Get dependencies, throws ExtensionError
		actions = objects.get<Diorite.Actions>("actions");
		storage = objects.get<Diorite.Storage>("storage");
		main_window = objects.get<MainWindow>("main_window");
		
		actions.add_action(new Diorite.NormalAction("debugging-audio", null, _("Audio Support"), null, null, open_audio_test));
		actions.add_action(new Diorite.NormalAction("debugging-flash", null, _("Flash Support"), null, null, open_flash_test));
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void add_actions_ui(Gtk.UIManager manager){
		debug("[%s] add_ui", id);
		try{
			if(main_window.mode == UIActionsMode.TOOLBAR){
				ui_merge_id = manager.add_ui_from_string(
					UI_SPEC_TOOLBAR, UI_SPEC_TOOLBAR.length);
			}
			else{
				ui_merge_id = manager.add_ui_from_string(
					UI_SPEC, UI_SPEC.length);
			}
		}
		catch(GLib.Error e){
			warning("[%s] Unable to add ui: %s", id, e.message);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void remove_actions_ui(Gtk.UIManager manager){
		debug("[%s] remove_ui", id);
		if(ui_merge_id != 0){
			manager.remove_ui(ui_merge_id);
			ui_merge_id = 0;
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		debug("[%s] unload", id);
	}
	
	private void open_audio_test(){
		if(audio_test != null){
			audio_test.present();
			return;
		}
		audio_test = new Gtk.Window();
		audio_test.set_default_size(550, 450);
		var test = storage.get_data_file("html/Audio.html");
		var view = new StaticWebView(test);
		var scroll = new Gtk.ScrolledWindow(null, null);
		scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
		scroll.add(view);
		audio_test.add(scroll);
		audio_test.delete_event.connect(on_audio_delete_event);
		audio_test.show_all();
		
	}
	
	private bool on_audio_delete_event(Gdk.EventAny event){
		audio_test.destroy();
		audio_test = null;
		return true;
	}
	
	private void open_flash_test(){
		if(flash_test != null){
			flash_test.present();
			return;
		}
		flash_test = new Gtk.Window();
		flash_test.set_default_size(550, 450);
		var test = storage.get_data_file("html/Flash.html");
		var view = new StaticWebView(test, true);
		var scroll = new Gtk.ScrolledWindow(null, null);
		scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC);
		scroll.add(view);
		flash_test.add(scroll);
		flash_test.delete_event.connect(on_flash_delete_event);
		flash_test.show_all();
		
	}
	
	private bool on_flash_delete_event(Gdk.EventAny event){
		flash_test.destroy();
		flash_test = null;
		return true;
	}
}

const string UI_SPEC = """
<ui>
	<menubar action="menubar">
		<menu action="service">
			<placeholder name="extras">
				<menuitem action="debugging-audio" />
				<menuitem action="debugging-flash" />
			</placeholder>
		</menu>
	</menubar>
</ui>
""";

const string UI_SPEC_TOOLBAR = """
<ui>
	<menubar action="menubar">
		<menu action="nuvola">
			<placeholder name="extras">
				<menuitem action="debugging-audio" />
				<menuitem action="debugging-flash" />
			</placeholder>
		</menu>
	</menubar>
</ui>
""";

} // namespace Nuvola.Extensions.Sample

