/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if LASTFM

using Diorite.Logger;

private int64 get_timestamp(){
	#if HAVE_GLIB_2_26
		var datetime = new DateTime.now_utc();
		return datetime.to_unix();
	#else
		// Deprecated
		TimeVal timeval = TimeVal();
		return (int64) timeval.tv_sec;
	#endif
}

/**
 * Module providing scrobbling to [[http://www.last.fm/|Last FM]]
 */
namespace Nuvola.Extensions.Lastfm{
	
/**
 * Last FM integration errors
 * 
 * * JSON_PARSE_ERROR - failed to parse Last FM response
 * * LASTFM_ERROR - invalid Last FM request
 * * WRONG_RESPONSE - invalid Last FM response
 */
public errordomain Error{
	JSON_PARSE_ERROR,
	LASTFM_ERROR,
	WRONG_RESPONSE
}

/**
 * HTTP method used to send request
 */
private enum Method{
	GET,
	POST;
}

/**
 * Last FM API key identifying Nuvola Player. You have to obtain your own
 * for your application, see http://www.last.fm/api/account
 */
private const string API_KEY = "209d19abf80aa2c6d049216ce46830d7";
/**
 * Last FM secret used to sign requests. You have to obtain your own
 * for your application, see http://www.last.fm/api/account
 */
private const string API_SECRET = "d157730073941bdd851eac950f3154e6";
/**
 * Root URL of Last FM API
 */
private const string API_ROOT = "http://ws.audioscrobbler.com/2.0/";

/**
 * Delay before song is scrobbled.
 */
private const int SCROBBLE_SONG_DELAY = 60;
/**
 * Delay before "now playing" status is updated.
 */
private const int NOW_PLAYING_TIMEOUT = 20;

/**
 * Obtains request token
 * 
 * This is the first step of desktop application authentication.
 * See http://www.last.fm/api/desktopauth for details.
 * 
 * @return request token
 * @throws Error on failure
 */
private async string get_request_token() throws Error{
	// http://www.last.fm/api/show/auth.getToken
	var params = new HashTable<string,string>(null, null);
	params.insert("method", "auth.getToken");
	params.insert("api_key", API_KEY);
	
	var response = yield send_request(Method.GET, params);
	if(!response.has_member("token")){
		throw new Error.WRONG_RESPONSE("Response doesn't contain token member.");
	}
	
	return response.get_string_member("token");
}

/**
 * Generates URL of authorization page
 * 
 * This is the second step of desktop application authentication. User has to
 * authorize provided request token. 
 * See http://www.last.fm/api/desktopauth for details.
 * 
 * @param token request token obtained by get_request_token()
 * @return url of the authorization page
 */
private string generate_auth_url(string token){
	return "http://www.last.fm/api/auth/?api_key=%s&token=%s".printf(
		 API_KEY, token);
}

/**
 * Obtains session key
 * 
 * This is the third step of desktop application authentication. Session key is used
 * to authorize Last FM API requests.
 * See http://www.last.fm/api/desktopauth for details.
 * 
 * @param request_token request token obtained by get_request_token()
 * @return session token
 * @throws Error on failure
 */
private async string get_session_key(string request_token) throws Error{
	// http://www.last.fm/api/show/auth.getSession
	var params = new HashTable<string,string>(null, null);
	params.insert("method", "auth.getSession");
	params.insert("api_key", API_KEY);
	params.insert("token", request_token);
	
	var response = yield send_request(Method.GET, params);
	if(!response.has_member("session")){
		throw new Error.WRONG_RESPONSE("Response doesn't contain session member.");
	}
	
	var session = response.get_object_member("session");
	if(!session.has_member("key")){
		throw new Error.WRONG_RESPONSE("Response doesn't contain session.key member.");
	}
	
	return session.get_string_member("key");
}

/**
 * Send Last FM API request
 * 
 * @param method HTTP method to use to send request
 * @param params Last FM API parameters of request
 * @return Root JSON object of the response
 * @throws Error on failure
 */
private async Json.Object send_request(Method method, HashTable<string,string> params) throws Error{
	Soup.Message message;
	var request = create_signed_request(params) + "&format=json";
	if(method == Method.GET){
		message = new Soup.Message("GET", API_ROOT + "?" + request);
	}
	else if(method == Method.POST){
		message = new Soup.Message("POST", API_ROOT);
		message.set_request("application/x-www-form-urlencoded",
			Soup.MemoryUse.COPY, request.data);
	}
	else{
		message = null;
		error("Last FM: Unsupported request method: %d", method);
	}
			
	var session = WebKit.get_default_session();
	SourceFunc callback = send_request.callback;
	session.queue_message(message, () => {
		Idle.add((owned) callback);
	});
	yield;
	
	var response = (string) message.response_body.flatten().data;
//~ 		debug("Response: %s", response);
	
	var parser = new Json.Parser();
	try{
		parser.load_from_data(response);
	}
	catch(GLib.Error e){
		throw new Error.JSON_PARSE_ERROR(e.message);
	}
	
	var root_object = parser.get_root().get_object();
	if(root_object.has_member("error") && root_object.has_member("message")){
		throw new Error.LASTFM_ERROR("%s: %s".printf(
			root_object.get_int_member("error").to_string(),
			root_object.get_string_member("message")
			));
	}
	return root_object;
}

/**
 * Creates signed request string for Last FM API call
 * 
 * @param params parameters of the request
 */
private string create_signed_request(HashTable<string,string> params){
	// See http://www.last.fm/api/desktopauth#6
	
	// Buffer for request string
	var req_buffer = new StringBuilder();
	// Buffer to compute signature for request
	var sig_buffer = new StringBuilder();
	
	// Signature requires sorted params
	var keys = params.get_keys();
	keys.sort(strcmp);
			
	foreach(var key in keys){
		var val = params.lookup(key);
		// signature buffer does not contain "=" and "&"
		// to separate key and value or key-value pairs
		// TODO: how about escaping?
		sig_buffer.append(key);
		sig_buffer.append(val);
		
		// request buffer contains "=" and "&"
		// to separate key and value or key-value pairs
		append_param(req_buffer, key, val);
	}
	
	// Append API_SECRET and generate MD5 hash
	sig_buffer.append(API_SECRET);
//~ 		debug("sig buffer: %s", sig_buffer.str);
	var api_sig = Checksum.compute_for_string(ChecksumType.MD5, sig_buffer.str);
	sig_buffer.truncate();
	
	// Append signature to the request string
	append_param(req_buffer, "api_sig", api_sig);
//~ 		debug("req buffer: %s", req_buffer.str);
	return req_buffer.str;
}

/**
 * Appends URL parameter
 * 
 * Appends URL parameter in format "key=value" or "&key=value".
 * 
 * @param buffer Buffer which parameter will be appended to
 * @param key parameter name
 * @param value parameter value
 */
public void append_param(StringBuilder buffer, string key, string value){
	if(buffer.len > 0){
		buffer.append_c('&');
	}
	buffer.append(Uri.escape_string(key, "", true));
	buffer.append_c('=');
	buffer.append(Uri.escape_string(value, "", true)); 
}

/**
 * Configuration key to store session key
 */
private const string SESSION_KEY = EXTENSIONS_KEY + ".lastfm.session_key";

/**
 * Last FM client
 */
public class Extension: Nuvola.Extension{
	private weak Diorite.MultiTypeMap config;
	private weak Nuvola.Player player;
	private Gtk.Box preferences;
	private uint scrobble_track_source = 0;
	private uint now_playing_source = 0;
	private string? token = null;
	private string? song = null;
	private string? artist = null;
	private int64 timestamp = 0;
	private bool can_scrobble = false;
	private string? _session;
	private string? session{
		get{
			return _session;
		}
		set{
			_session = value;
			config[SESSION_KEY] = _session ?? "";
			var serializable = config as Diorite.Serializable;
			if(serializable != null){
				try{
					serializable.save();
				}
				catch(GLib.Error e){
					warning("Unable to save configuration: %s", e.message);
				}
			}
		}
	}
	
	construct{
		has_preferences = true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		player.song_changed.disconnect(on_song_changed);
		var timestamp = get_timestamp();
		if(can_scrobble && this.song != null && this.artist != null){
			scrobble_track(song, artist, timestamp, (obj, res) => {
				try{
					scrobble_track.end(res);
				}
				catch(Error e){
					warning("Last FM error: %s", e.message);
				}
			});
		}
		if(scrobble_track_source != 0){
			Source.remove(scrobble_track_source);
			scrobble_track_source = 0;
		}
		
		if(now_playing_source != 0){
			Source.remove(now_playing_source);
			now_playing_source = 0;
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override Gtk.Widget? get_preferences(){
		preferences = new Gtk.VBox(false, 2);
		update_preferences();
		return preferences;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		config = objects.get<Diorite.MultiTypeMap>("config");
		player = objects.get<Player>("player");
		_session = config[SESSION_KEY];
		if(_session == ""){
			_session = null;
		}
		player.song_changed.connect(on_song_changed);
	}
	
	/**
	 * Whether client has request token to obtain session key
	 * @return true if client has request token
	 */
	public bool has_token(){
		return token != null;
	}
	
	/**
	 * Whether client has session key to authorize requests
	 * @return true if client has session key
	 */
	public bool has_session(){
		return _session != null;
	}
	
	/**
	 * Generates authorization URL to authorize request token
	 * 
	 * @return authorization URL
	 * @throws Error on failure
	 */
	public async string get_auth_url() throws Error{
		token = yield get_request_token();
		return generate_auth_url(token);
	}
	
	/**
	 * Exchanges authorized request token for session key.
	 * 
	 * @throws Error on failure
	 */
	public async void fetch_session() throws Error{
		session = yield get_session_key(token);
		token = null;
	}
	
	/**
	 * Forgets session key
	 */
	public void drop_session(){
		session = null;
	}
	
	/**
	 * Forgets request token
	 */
	public void drop_token(){
		token = null;
	}
	
	/**
	 * Callback for song change. Queues "now playing" update and track scrobbling.
	 */
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		var timestamp = get_timestamp();
		
		if(scrobble_track_source != 0){
			Source.remove(scrobble_track_source);
			scrobble_track_source = 0;
		}
		
		if(now_playing_source != 0){
			Source.remove(now_playing_source);
			now_playing_source = 0;
		}
		
		if(session == null
		 ||song == null || song == ""
		 || artist == null || artist == ""){
			return;
		}
		
		if(can_scrobble && this.song != null && this.artist != null){
			scrobble_track(this.song, this.artist, this.timestamp, (obj, res) => {
				try{
					scrobble_track.end(res);
				}
				catch(Error e){
					warning("Last FM error: %s", e.message);
				}
			});
		}
		
		this.song = song;
		this.artist = artist;
		this.timestamp = timestamp;
		can_scrobble = false;
		
		
		update_now_playing(song, artist, (obj, res) => {
			try {
				update_now_playing.end(res);
			}
			catch(Error e){
				warning("Last FM error: %s", e.message);
			}
		});
		
		scrobble_track_source = Timeout.add_seconds(SCROBBLE_SONG_DELAY, enable_scrobbling_cb);
		now_playing_source = Timeout.add_seconds(NOW_PLAYING_TIMEOUT, update_now_playing_cb);
	}
	
	private bool enable_scrobbling_cb(){
		if(scrobble_track_source != 0){
			can_scrobble = true;
			scrobble_track_source = 0;
		}
		
		return false;
	}
	
	/**
	 * Scrobbles track to Last FM
	 * 
	 * @param song song name
	 * @param artist artist name
	 * @param timestamp Unix time
	 * @throws Error on failure
	 */
	private async void scrobble_track(string song, string artist, int64 timestamp) throws Error{
		debug("Last FM scrobble: %s by %s, %s", song, artist, timestamp.to_string());
		if(session == null) return;
		
		// http://www.last.fm/api/show/track.scrobble
		var params = new HashTable<string,string>(null, null);
		params.insert("method", "track.scrobble");
		params.insert("api_key", API_KEY);
		params.insert("sk", session);
		params.insert("track", song);
		params.insert("artist", artist);
		params.insert("timestamp", timestamp.to_string());
	
		var response = yield send_request(Method.POST, params);
		if(!response.has_member("scrobbles")){
			throw new Error.WRONG_RESPONSE("Response doesn't contain scrobbles member.");
		}
	}
	
	private bool update_now_playing_cb(){
		if(now_playing_source == 0 || song == null || artist == null) return false;
		
		update_now_playing(song, artist, (obj, res) => {
			try {
				update_now_playing.end(res);
			}
			catch(Error e){
				warning("Last FM error: %s", e.message);
			}
		});
		return true;
	}
	
	/**
	 * Updates now playing status on Last FM
	 * 
	 * @param song song name
	 * @param artist artist name
	 * @throws Error on failure
	 */
	private async void update_now_playing(string song, string artist) throws Error{
		debug("Last FM update now playing: %s by %s", song, artist);
		if(session == null || player.artist != artist || player.song != song) return;
		
		// http://www.last.fm/api/show/track.updateNowPlaying
		var params = new HashTable<string,string>(null, null);
		params.insert("method", "track.updateNowPlaying");
		params.insert("api_key", API_KEY);
		params.insert("sk", session);
		params.insert("track", song);
		params.insert("artist", artist);
	
		var response = yield send_request(Method.POST, params);
		if(!response.has_member("nowplaying")){
			throw new Error.WRONG_RESPONSE("Response doesn't contain nowplaying member.");
		}
	}
	
	private void update_preferences(){
		foreach(var child in preferences.get_children()){
			preferences.remove(child);
		}
		
		Gtk.Label label;
		Gtk.Button button;
		
		if(has_session()){
			label = new Gtk.Label(_("Scrobbling to Last FM is enabled."));
			preferences.pack_start(label, false, false, 10);
			button = new Gtk.Button.with_label(_("Unauthorize"));
			button.clicked.connect(() => {
				drop_session();
				update_preferences();
			});
			preferences.pack_start(button, false, false, 10);
		}
		else if(has_token()){
			label = new Gtk.Label(_("Last FM authorization page should have been opened in your web browser. Allow acces to Nuvola Player and click on the button bellow to finish authorization process."));
			preferences.pack_start(label, false, false, 10);
			button = new Gtk.Button.with_label(_("Complete authorization"));
			button.clicked.connect(() => {
				button.sensitive = false;
				fetch_session((obj, res) => {
					try{
						fetch_session.end(res);
						update_preferences();
					}
					catch(Lastfm.Error e){
						warning("Last FM error: %s", e.message);
						button.label = _("Error occurred. Try again.");
					}
					button.sensitive = true;
				});
			});
			preferences.pack_start(button, false, false, 10);
		}
		else{
			label = new Gtk.Label(_("You have to grant access to Nuvola Player to enable scrobbling to Last FM. Authorization page will be opened in your web browser."));
			preferences.pack_start(label, false, false, 10);
			button = new Gtk.Button.with_label(_("Authorize"));
			button.clicked.connect(() => {
				button.sensitive = false;
				get_auth_url((obj, res) => {
					try{
						var url = get_auth_url.end(res);
						Diorite.System.open_uri_with_fallback(url);
						update_preferences();
					}
					catch(Lastfm.Error e){
						warning("Last FM error: %s", e.message);
						button.label = _("Error occurred. Try again.");
					}
					button.sensitive = true;
				});
			});
			preferences.pack_start(button, false, false, 10);
		}
		
		label.xpad = 0;
		label.use_markup = true;
		label.wrap_mode = Pango.WrapMode.WORD;
		label.wrap = true;
		label.justify = Gtk.Justification.FILL;
			
		preferences.show_all();
	}
}

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing scrobbling to Last.fm
		_("Last FM scrobbling"),
		Nuvola.VERSION,
		/// Extension descriptiom
		_("This plugin provides scrobbling to Last FM (http://last.fm)"),
		"Jiří Janoušek",
		typeof(Extension)
	};
}

} // namespace Nuvola.Extensions.Lastfm

#endif
