/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if NOTIFICATIONS

using Diorite.Logger;
using Diorite;
using Nuvola.Settings;

namespace Nuvola.Extensions.Notifications{

const string CONF_ID = "notifications";
const string ACTIVE_WINDOW = EXTENSIONS_KEY + "." + CONF_ID + "." + "active_window";
const string RESIDENT = EXTENSIONS_KEY + "." + CONF_ID + "." + "resident";

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing integration with multimedia keys in GNOME
		_("Notifications"),
		Nuvola.VERSION,
		/// Description of a plugin providing integration with multimedia keys in GNOME
		_("This plugin provides desktop notifications (libnotify)."),
		"Jiří Janoušek",
		typeof(Extension)
	};
}

/**
 * Manages notifications
 */
public class Extension: Nuvola.Extension{
	private weak Diorite.Application app;
	private weak Player player;
	private weak MainWindow main_window;
	private weak MultiTypeMap config;
	private Notify.Notification? notification = null;
	
	public bool active_window{
		get{
			return config.get_bool(ACTIVE_WINDOW, false);
		}
		set{
			config.set_bool(ACTIVE_WINDOW, value);
		}
	}
	
	#if FALSE && EXPERIMENTAL && LIBNOTIFY_7
	public bool resident{
		get{
			return config.get_bool(RESIDENT, true);
		}
		set{
			config.set_bool(RESIDENT, value);
			toggle_resident(value);
		}
	}
	#endif
	
	construct{
		has_preferences = true;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		app = objects.get<Diorite.Application>("application");
		player = objects.get<Player>("player");
		main_window = objects.get<MainWindow>("main_window");
		config = objects.get<Diorite.MultiTypeMap>("config");
		
		Notify.init(app.display_name);
		/// Initial notification
		notification = new Notify.Notification(app.display_name, _("Version") + " " + app.version, app.icon);
		#if FALSE && EXPERIMENTAL && LIBNOTIFY_7
		toggle_resident(resident);
		#endif
		
		player.song_changed.connect(on_song_changed);
		player.notify["playback-state"].connect(on_state_changed);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		player.notify["playback-state"].disconnect(on_state_changed);
		player.song_changed.disconnect(on_song_changed);
		#if FALSE && EXPERIMENTAL && LIBNOTIFY_7
		if(resident){
			player.notify["actions"].disconnect(on_actions_changed);
		}
		#endif
		Notify.uninit();
		notification = null;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override Gtk.Widget? get_preferences(){
		var box = new Gtk.VBox(false, 0);
		/// Settings dialog - tab General
		var active_window = new Gtk.CheckButton.with_label(_("Notify also when window is active."));
		active_window.active = this.active_window;
		active_window.toggled.connect(() => { this.active_window = active_window.active; });
		box.pack_start(active_window, false, false);
		
		#if FALSE && EXPERIMENTAL && LIBNOTIFY_7
		/// Settings dialog - tab General
		var resident = new Gtk.CheckButton.with_label(_("Enable resident notifications"));
		resident.active = this.resident;
		resident.toggled.connect(() => { this.resident = resident.active; });
		box.pack_start(resident, false, false);
		#endif
		
		box.show_all();
		return box;
	}
	
	private void on_state_changed(Object o, ParamSpec p){
		show_notification();
	}
	
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		if(song == null || song == "") return;
		/// Notifications
		string text;
		if(artist == null && album == null){
			/// Shown in notification if both artist and album are unknown
			text = _("by unknown artist");
		}
		else if(artist == null){
			/// Shown in notification if artist is unknown, %s will be replaced by album name
			text = _("by unknown artist from %s").printf(album);
		}
		else if(album == null){
			/// Shown in notification if album is unknown, %s will be replaced by artist name
			text = _("by %s").printf(artist);
		}
		else{
			/// Shown in notification, %1$s will be replaced by artist name, %2$s will be replaced by album name
			text = _("by %1$s from %2$s").printf(artist, album);
		}
		this.notification.update(song, text, album_art ?? "");
		show_notification();
	}
	
	private void show_notification(){
		return_if_fail(notification != null);
		if(player.playback_state != STATE_PLAYING) return;
		if(main_window.is_active && !active_window) return;
		try{
				this.notification.show();
			}
			catch(Error e){
				warning("Unable to show notification: %s", e.message);
			}
		
	}
	
	#if FALSE && EXPERIMENTAL && LIBNOTIFY_7
	private void toggle_resident(bool resident){
		return_if_fail(notification != null);
		if(resident){
			notification.set_hint("resident", true);
			notification.set_hint("action-icons", true);
			player.notify["actions"].connect_after(on_actions_changed);
			
			
		}
		else{
			notification.set_hint("resident", false);
			notification.set_hint("action-icons", false);
			notification.clear_actions();
			player.notify["actions"].disconnect(on_actions_changed);
		}
	}
	
	private void on_actions_changed(Object o, ParamSpec p){
		update_actions();
	}
	
	private void update_actions(){
		return_if_fail(notification != null);
		
		notification.clear_actions();
		// TODO: notification.clear_actions() seems not to work. Why?
		// Remove EXPERIMENTAL when fixed
		
		if((player.actions & PlayerActions.PREV_SONG) != 0){
			notification.add_action(PlayerActions.PREV_SONG.get_icon(), PlayerActions.PREV_SONG.get_label(),
				() => { player.previous_song(); });
		}
		if((player.actions & PlayerActions.PLAY) != 0){
			notification.add_action(PlayerActions.PLAY.get_icon(), PlayerActions.PLAY.get_label(),
				() => { player.play(); });
		}
		if((player.actions & PlayerActions.PAUSE) != 0){
			notification.add_action(PlayerActions.PAUSE.get_icon(), PlayerActions.PAUSE.get_label(),
				() => { player.pause(); });
		}
		if((player.actions & PlayerActions.NEXT_SONG) != 0){
			notification.add_action(PlayerActions.NEXT_SONG.get_icon(), PlayerActions.NEXT_SONG.get_label(),
				() => { player.next_song(); });
		}
	}
	#endif
}

} // namespace Nuvola.Extensions.Notifications
#endif
