/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Nuvola.Extensions.TrayIcon{

public Nuvola.ExtensionInfo get_info(){
	return {
		/// Name of a plugin providing scrobbling to Last.fm
		_("Tray Icon"),
		Nuvola.VERSION,
		/// Extension descriptiom
		_("This plugin shows tray icon with menu. Tray icon may be required for 'hide on close' feature."),
		"Jiří Janoušek",
		typeof(Extension)
	};
}

/**
 * Tray icon is used to control playback when the main window is not visible
 * (it's minimized, hidden or covered by other windows) and to bring the hidden main
 * window to foreground.
 * 
 * The tray icon is shown only if Ubuntu Sound menu is not detected.
 */
public class Extension: Nuvola.Extension{
	private weak Diorite.Application app;
	private weak Diorite.Actions actions;
	private weak Nuvola.MainWindow main_window;
	private weak Nuvola.Player player;
	private Gtk.Menu? menu;
	private Gtk.StatusIcon? icon;
	private Gtk.Widget[] extra_actions = {};
	private uint update_id = 0;
	
	/**
	 * {@inheritDoc}
	 */
	public override void load(ObjectContainer objects) throws ExtensionError{
		// Get dependencies, throws ExtensionError
		app = objects.get<Diorite.Application>("application");
		main_window = objects.get<Nuvola.MainWindow>("main_window");
		actions = objects.get<Diorite.Actions>("actions");
		player = objects.get<Nuvola.Player>("player");
		
		icon = new Gtk.StatusIcon.from_icon_name(APP_ICON);
		icon.title = app.display_name;
		update_tooltip();
		icon.popup_menu.connect(show_menu);
		icon.activate.connect(() => {
			app.activate();
		});
		
		menu = new Gtk.Menu();
		foreach(var name in PLAYER_ACTIONS){
			var action = actions.get_action(name);
			if(action != null){
				menu.append(create_menu_item(action));
			}
			else{
				critical("Action '%s' not found.", name);
			}
		}
		
		var quit_action = actions.get_action("quit");
		if(quit_action != null){
			menu.append(create_menu_item(quit_action));
		}
		else{
			critical("Action '%s' not found.", "quit");
		}
		
		foreach(var name in PLAYER_ACTIONS_EXTRA){
			var action = actions.get_action(name);
			if(action == null){
				critical("Action '%s' not found.", name);
			}
			else{
				action.notify["sensitive"].connect(on_actions_changed);
			}
		}
		update_menu();
		main_window.close_action.connect(on_close_action);
		player.song_changed.connect(on_song_changed);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public override void unload(){
		foreach(var name in PLAYER_ACTIONS_EXTRA){
			var action = actions.get_action(name);
			if(action == null){
				critical("Action '%s' not found.", name);
			}
			else{
				action.notify["sensitive"].disconnect(on_actions_changed);
			}
		}
		main_window.close_action.disconnect(on_close_action);
		player.song_changed.disconnect(on_song_changed);
		icon.visible = false;
		icon = null;
		menu = null;
	}
	
	private void on_actions_changed(Object o, ParamSpec p){
		lock(update_id){
			if(update_id == 0){
				update_id = Idle.add(update_menu_cb);
			}
		}
	}
	
	private void on_close_action(ref bool quit){
		if(icon != null && icon.visible){
			quit = false;
		}
	}
	
	private bool update_menu_cb(){
		update_menu();
		return false;
	}
	
	private void update_menu(){
		lock(update_id){
			update_id = 0;
		}
		return_if_fail(menu != null);
		
		foreach(var item in extra_actions){
			menu.remove(item);
		}
		
		extra_actions = {};
		int index = PLAYER_ACTIONS.length;
		foreach(var name in PLAYER_ACTIONS_EXTRA){
			var action = actions.get_action(name);
			if(action == null){
				critical("Action '%s' not found.", name);
			}
			else if(action.sensitive){
				var menu_item = create_menu_item(action);
				menu.insert(menu_item, index++);
				extra_actions += menu_item;
			}
		}
	}
	
	private Gtk.MenuItem create_menu_item(Diorite.Action action){
		Gtk.MenuItem item;
		if(action.stock_id != null){
			var image_item = new Gtk.ImageMenuItem();
			image_item.image = new Gtk.Image.from_icon_name(action.stock_id, Gtk.IconSize.MENU);
			
			action.notify["stock-id"].connect((o, p) => {
				if(action.stock_id != null){
					image_item.image = new Gtk.Image.from_icon_name(action.stock_id, Gtk.IconSize.MENU);
				}
			});
			item = image_item;
		}
		else{
			item = new Gtk.MenuItem();
		}
		
		item.label = action.label.replace("_", "");
		action.notify["label"].connect((o, p) => {
			item.label = action.label.replace("_", "");
		});
		
		item.sensitive = action.sensitive;
		action.notify["sensitive"].connect((o, p) => {
			item.sensitive = action.sensitive;
		});
		item.activate.connect(() => { action.activate(); });
		return item;
	}
	
	private void show_menu(uint button, uint time){
		return_if_fail(menu != null);
		menu.show_all();
		menu.popup(null, null, icon.position_menu, button, time);
	}
	
	private void on_song_changed(string? song, string? artist, string? album, string? album_art){
		update_tooltip();
	}
	
	private void update_tooltip(){
		if(player.playback_state == STATE_NONE || player.song == null){
			icon.tooltip_text = app.display_name;
		}
		else if(player.artist == null){
			icon.tooltip_text = "%s - %s".printf(player.song, app.display_name);
		}
		else{
			icon.tooltip_text = "%s - %s - %s".printf(player.song, player.artist, app.display_name);
		}
	}
}

} // namespace Nuvola.Extensions.TrayIcon
