/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using Diorite.Logger;

namespace Nuvola{

/**
 * ExtensionManagerView provided graphicas interface for extensions management.
 */
public class ExtensionsManagerView: Gtk.HBox{
	private Gtk.TreeView selector;
	private Gtk.ListStore model;
	private weak ExtensionsManager extensions;
	private weak Diorite.MultiTypeMap config;
	private Gtk.TextView description;
	private Gtk.Widget? preferences;
	private Gtk.Button settings_button;
	
	/**
	 * Creates new ExtensionsManagerView
	 * 
	 * @param extensions	extensions manager
	 * @param config		configuration
	 */
	public ExtensionsManagerView(ExtensionsManager extensions, Diorite.MultiTypeMap config){
		this.extensions = extensions;
		this.config = config;
		
		// id, enabled, name
		model = new Gtk.ListStore(EMI.COUNT, typeof(string), typeof(bool), typeof(string));
		var keys = extensions.available_extensions.get_keys();
		keys.sort(strcmp);
		var enabled = config.array(EXTENSIONS_KEY);
		Gtk.TreeIter iter;
		foreach(var key in keys){
			var extension = extensions.available_extensions.lookup(key);
			model.append(out iter);
			model.set(iter, EMI.ID, key, EMI.ACTIVE, (key in enabled), EMI.NAME, extension.name);
		}
		
		var win = new Gtk.ScrolledWindow(null, null);
		selector = new Gtk.TreeView.with_model(model);
		selector.headers_clickable = true;
		win.add(selector);
		win.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		win.shadow_type = Gtk.ShadowType.IN;
		border_width = 3;
		win.border_width = 3;
		var box = new Gtk.VBox(false, 0);
		box.pack_start(win, true, true, 0);
		pack_start(box, false, false, 0);
		
		
		var enabled_renderer = new Gtk.CellRendererToggle();
		enabled_renderer.toggled.connect(on_extension_toggled);
		var enabled_column = new Gtk.TreeViewColumn.with_attributes(
			/// header in extensions manager
				_("Enabled"), enabled_renderer, "active", EMI.ACTIVE);
		var name_renderer = new Gtk.CellRendererText();
		var name_column = new Gtk.TreeViewColumn.with_attributes(
			/// header in extensions manager
				_("Extension name"), name_renderer, "text", EMI.NAME);
		name_column.sort_column_id = EMI.NAME;
		selector.append_column(enabled_column);
		selector.append_column(name_column);
		
		
		var vbox = new Gtk.VBox(false,0 );
		pack_start(vbox, true, true, 0);
		description = new Gtk.TextView();
		description.editable = false;
		description.wrap_mode = Gtk.WrapMode.WORD_CHAR;
		
		var scroll_box = new Gtk.ScrolledWindow(null, null);;
		scroll_box.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
		win.shadow_type = Gtk.ShadowType.IN;
		vbox.pack_start(scroll_box, true, true, 0);
		
		scroll_box.add(description);
		scroll_box.set_size_request(200, -1);
		
		settings_button = new Gtk.Button.from_stock(Gtk.Stock.PREFERENCES);
		vbox.pack_start(settings_button, false, false, 0);
		settings_button.sensitive = false;
		settings_button.clicked.connect(on_settings_button_clicked);
		
		var selection = selector.get_selection();
		selection.changed.connect(on_selection_changed);
		selection.set_mode(Gtk.SelectionMode.SINGLE);
		if(model.get_iter_first(out iter)){
			selection.select_iter(iter);
		}
		selector.grab_focus();
	}
	
	/**
	 * Called when extension is selected.
	 */
	private void on_selection_changed(Gtk.TreeSelection selection){
		Gtk.TreeIter iter;
		string id;
		bool enabled;
		if(selection.get_selected(null, out iter)){
			model.get(iter, EMI.ID, out id, EMI.ACTIVE, out enabled);
			show_info(id, enabled);
		}
	}
	
	/**
	 * Shows information about extension
	 *
	 * @param id			id of extension
	 * @param enabled		whether the extension is currently enabled
	 */
	private void show_info(string id, bool enabled){
		var info = extensions.available_extensions.lookup(id);
		
		/// label for name of extension
		var text = "%s: %s\n%s: %s\n%s: %s\n\n%s".printf(_("Name"), info.name, 
			/// label for version of extension
			_("Version"), info.version,
			/// label for author of extension
			_("Author"), info.author, info.description);
		description.get_buffer().set_text(text);
		settings_button.sensitive = false;
		
		preferences = null;
		
		if(enabled){
			var extension = extensions[id];
			if(extension != null){
				settings_button.sensitive = extension.has_preferences;
				preferences = extension.get_preferences();
			}
		}
	}
	
	private void on_settings_button_clicked(){
		var dialog = new Gtk.Dialog.with_buttons(_("Extension settings"),
			get_toplevel() as Gtk.Window, Gtk.DialogFlags.MODAL,
			Gtk.Stock.CLOSE, 1);
		var box = dialog.get_content_area() as Gtk.Box;
		dialog.border_width = 10;
		box.add(preferences);
		box.show_all();
		dialog.run();
		box.remove(preferences);
		dialog.destroy();
	}
	
	/**
	 * Called when extension is turned on/off.
	 */
	private void on_extension_toggled(string path){
		Gtk.TreeIter iter;
		var treepath =  new Gtk.TreePath.from_string(path);
		string id;
		bool enabled;
		if(model.get_iter( out iter, treepath)){
			model.get(iter, EMI.ID, out id, EMI.ACTIVE, out enabled);
			if(enabled){
				if(extensions.unload(id)){
					enabled = !enabled;
				}
			}
			else{
				if(extensions.load(id) != null){
					enabled = !enabled;
				}
			}
			model.set(iter, EMI.ACTIVE, enabled);
			show_info(id, enabled);
			if(model.get_iter_first(out iter)){
				string[] enabled_extensions = {};
				do{
					model.get(iter, EMI.ID, out id, EMI.ACTIVE, out enabled);
					if(enabled){
						enabled_extensions += id;
					}
				}
				while(model.iter_next(ref iter));
				config.set_array(EXTENSIONS_KEY, enabled_extensions);
			}
		}
	}
}


private enum EMI{
	ID,
	ACTIVE,
	NAME,
	COUNT;
}

} // namespace Nuvola
