/*
 * Copyright 2011-2012 Jiří Janoušek <janousek.jiri@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using WebKit;
using Soup;
using Gee;
using Diorite;
using Diorite.Logger;
using Diorite.Widgets;
using Diorite.System;
using JS;
using Nuvola.JSTools;

/**
 * Frame with the web interface of cloud player service.
 * Handles state changes received from WebView and produced by script
 * loaded in WebView, console logging, WebView requests, user scripts management, etc.
 */
public class Nuvola.MainWebView: WebView{
	private SList<string> blocked_resources;
	private weak Diorite.Actions? actions;
	private Diorite.Storage storage;
	private Service? _service = null;
	public Service? service{
		get{
			return _service;
		}
		set{
			set_maintains_back_forward_list(false);
			set_maintains_back_forward_list(true);
			_service = value;
		}
	}
	
	private bool _enable_navigation = true;
	public bool enable_navigation{
		get{
			return this._enable_navigation;
		}
		set{
			if(this._enable_navigation == value) return;
			this._enable_navigation = value;
			this.update_navigation();
		}
	}
	
	public bool web_plugins = false;
	
	public virtual signal void navigation_approved(string uri){
		debug_navigation("Navigation approved: %s", uri);
	}
	
	/**
	 * Constructs new Frame for application
	 * 
	 * @param app Application object
	 */
	public class MainWebView(Diorite.Storage storage, Diorite.Actions? actions=null){
		this.actions = actions;
		this.storage = storage;
		blocked_resources = new SList<string>();
		append_ui_actions();
		full_content_zoom = true;
		var settings =  this.get_settings();
		settings.enable_caret_browsing = true;
		settings.enable_plugins = !Args.no_web_plugins;
		settings.javascript_can_open_windows_automatically = true;
		settings.enable_site_specific_quirks  = true;
		settings.enable_default_context_menu = Args.debug_javascript;
//~ 			settings.user_agent = "Mozilla/5.0 (X11; Linux x86_64; rv:5.0) Gecko/20100101 Firefox/5.0 Firefox/5.0";
		settings.html5_local_storage_database_path = storage.user_config_dir.get_path();
		settings.enable_developer_extras = Args.debug_javascript;
		this.set_settings(settings);

		navigation_policy_decision_requested.connect(on_new_page);
		new_window_policy_decision_requested.connect(on_new_window);
		create_plugin_widget.connect(on_create_plugin_widget);
		resource_request_starting.connect(on_resource_request);
		
		if(Args.debug_javascript){
			new InspectorWindow(this);
		}
	}
	
	public override WebView create_web_view(WebFrame frame){
		/* Create new popup window */
		return new PopupWebView(this.get_settings());
	}

	/**
	 * Resets zoom level
	 */		
	public void zoom_reset(){
		this.zoom_level = 1.0f;
	}
	
	/**
	 * Loads home page (see home_page property)
	 */
	public void go_home(){
		if(service != null && service.home_page != null){
			open(service.home_page);
		}
	}
	
	private void on_load_status_changed(GLib.Object o, ParamSpec p){
		if(this._enable_navigation){
			actions.get_action("back").sensitive = can_go_back();
			actions.get_action("forward").sensitive = can_go_forward();
		}
	}
	
	private void update_navigation(){
		return_if_fail(actions != null);
		actions.get_action("refresh").sensitive = _enable_navigation;
		actions.get_action("home").sensitive = _enable_navigation;
		actions.get_action("back").sensitive = _enable_navigation && can_go_back();
		actions.get_action("forward").sensitive = _enable_navigation && can_go_forward();
	}
	
	/**
	 * Creates UI actions and appends them to UIActions instance
	 */
	private void append_ui_actions(){
		if(actions == null) return;
	
		/* Normal items. Name, stock id, label, accelerator, tooltip, callback */
		
		actions.add_actions({
			/// Tool bar item tooltip
			new NormalAction("home", Gtk.Stock.HOME, null, "<alt>Home", _("Back to main page of the service"), go_home),
			/// Tool bar item tooltip
			new NormalAction("refresh", Gtk.Stock.REFRESH, null, "F5", _("Reload current page"), reload),
			/// Tool bar item tooltip
			new NormalAction("back", Gtk.Stock.GO_BACK, null, null, _("Go to previous page"), go_back),
			/// Tool bar item tooltip
			new NormalAction("forward", Gtk.Stock.GO_FORWARD, null, null, _("Go to next page"), go_forward),
			new NormalAction("zoom-reset", Gtk.Stock.ZOOM_100, null, "<ctrl>0", null, zoom_reset),
			new NormalAction("zoom-in", Gtk.Stock.ZOOM_IN, null, "<ctrl>plus", null, zoom_in),
			new NormalAction("zoom-out", Gtk.Stock.ZOOM_OUT, null, "<ctrl>minus", null, zoom_out)
		});
		
		update_navigation();
		
		notify["load-status"].connect(on_load_status_changed);
		notify["uri"].connect(on_load_status_changed);
		notify["title"].connect(on_load_status_changed);
	}
	
	/**
	 * Decide whether open page in main view or in the default browser
	 */
	private bool on_new_page(WebFrame frame, NetworkRequest req,
		WebNavigationAction action, WebPolicyDecision decision
	){
		if(frame != get_main_frame()){
			// don't care about pages in <iframe> elements
			return false;
		}
		
		string uri = req.get_uri();
		debug_navigation("New page request: URI: %s, frame URI: %s",
			uri, frame.get_uri() ?? "null");
			
		
		/* Open Flash install page in default browser */
		if(uri.has_prefix("http://get.adobe.com/flashplayer")){
			open_uri_with_fallback(uri);
			decision.ignore();
			return true;
		}
		
		if(service != null){
			// If page doesn't match sandbox, open it in default browser
			if(!service.matches_sandbox(uri)){
				open_uri_with_fallback(uri);
				decision.ignore();
				return true;
			}
			else{
				navigation_approved(uri);
			}
		}
		return false;
	}
	
	/**
	 * Open new window in the default web browser
	 */
	private bool on_new_window(WebFrame frame, NetworkRequest req,
		WebNavigationAction action, WebPolicyDecision decision
	){
		unowned string uri = req.get_uri();
		debug_navigation("New window request: %s", uri);
		
		if(service != null && service.matches_sandbox(uri)){
			// Oh wait! This page belongs to the main wiew!
			open(uri);
		}
		else{
			open_uri_with_fallback(uri);
		}
		decision.ignore();
		return true;
	}
	
	/**
	 * Called when web plugin widget is requested. Appends URL to the black list.
	 */
	private unowned Gtk.Widget? on_create_plugin_widget(string mime, string uri, HashTable attributes){
		if(!web_plugins && service != null && uri != null && !service.matches_sandbox(uri)){
			debug("Widget blocked: %s", uri);
			blocked_resources.append(Uri.unescape_string(uri));
		}
		return null;
	}
	
	/**
	 * Called when web resource is requested. Block resources on black list.
	 */
	private void on_resource_request(WebKit.WebFrame frame, WebKit.WebResource resource,
		WebKit.NetworkRequest request, WebKit.NetworkResponse? response){
		var uri = request.uri;
		if(uri == null || web_plugins) return;
		uri = Uri.unescape_string(uri);
		unowned SList<string> item = blocked_resources.find_custom(uri, strcmp);
		if(item != null){
//~ 				debug("Request blocked: %s", uri);
			request.uri = "about:blank";
			blocked_resources.delete_link(item);
		}
	}
	
	/**
	 * Logs debug message with libfenryxo domain
	 * @param format format string
	 */
	private static void debug_navigation(string format, ...){
		if(Args.debug_navigation){
			va_list args = va_list();
			GLib.logv(Nuvola.LOG_DOMAIN, LogLevelFlags.LEVEL_DEBUG, format, args);
		}
	}
}
